import enum

from rest_framework import serializers
from rest_framework.fields import empty

from .mixins import WBCoreSerializerFieldMixin
from .types import WBCoreType


class ListFieldMixin:
    def get_attribute(self, instance):
        # Can't have any relationships if not created
        try:
            return super().get_attribute(instance)

        except (AttributeError, KeyError):
            return []

    def run_validation(self, data=empty):
        # If the data is send through form data, we need to convert the data into a proper list of ids
        if data not in [None, empty] and len(data) == 1 and isinstance(data[0], str) and "," in data[0]:
            data = data[0].split(",")

        # If the data is a list of an empty string we need to convert it (FORM DATA)
        if data not in [None, empty] and len(data) == 1 and isinstance(data[0], str) and data[0] == "":
            data = []

        # If the data is a list and contains the string null, then we need to convert it (FORM DATA)
        if data == ["null"]:
            data = []

        # If the data is None and null is an allowed value, data needs to be set to an empty list
        if data is None and self.allow_null:
            data = []
        return super().run_validation(data)


class ListField(ListFieldMixin, WBCoreSerializerFieldMixin, serializers.ListField):
    field_type = WBCoreType.LIST.value


class SparklineField(WBCoreSerializerFieldMixin, serializers.ListField):
    """Readonly field to create a sparline representation from an annotated aggregated array

    Important: This is potentially very slow, use accordingly

    Example:
        View:
            def get_queryset(self):
                return super().get_queryset().annotate(
                    prices_date=ArrayAgg("prices__date", filter=Q(prices__date__gt="2023-03-01")),
                    prices_price=ArrayAgg("prices__net_value", filter=Q(prices__date__gt="2023-03-01"))
                )

        Serializer:
            prices_sparkline = wb_serializers.SparklineField(x_data_label="prices_date", y_data_label="prices_price")
    """

    field_type = WBCoreType.SPARKLINE.value

    class Type(enum.Enum):
        LINE = "line"
        BAR = "bar"
        COLUMN = "column"
        AREA = "area"

    def __init__(
        self, *args, x_data_label=None, y_data_label=None, sparkline_type=Type.LINE, sparkline_option=None, **kwargs
    ):
        """
        Args:
            *args: based serializer positional argument
            x_data_label: The field name as find in the serialized object representative the easting data
            y_data_label: The field name as find in the serialized object representive the northing data
            sparkline_type: Either one of LINE, BAR, COLUMN or AREA
            sparkline_option: a json serializable dictionary representing additional option as allowed by ag-grid.
            **kwargs:based serializer keyword argument
        """
        self.x_data_label = x_data_label
        self.y_data_label = y_data_label
        self.sparkline_type = sparkline_type
        kwargs.pop("read_only", None)
        if not sparkline_option:
            sparkline_option = {}
        self.sparkline_option = sparkline_option
        super().__init__(*args, read_only=True, **kwargs)

    def get_attribute(self, obj):
        # We pass the object instance onto `to_representation`,
        # not just the field attribute.
        return obj

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["read_only"] = True
        representation["sparkline_type"] = self.sparkline_type.value
        representation["sparkline_option"] = self.sparkline_option
        return key, representation

    def to_representation(self, obj):
        representation = [[]]  # if row is [] or null, we default to an empty list of list
        if (x_data := getattr(obj, self.x_data_label, None)) and (y_data := getattr(obj, self.y_data_label, None)):
            representation = zip(x_data, y_data)
        return representation



from .base import WBCoreViewConfig


class PaginationViewConfig(WBCoreViewConfig):
    metadata_key = "pagination"
    config_class_attribute = "pagination_config_class"

    def get_metadata(self) -> str | None:
        if pagination := getattr(self.view, "pagination_class", None):
            return pagination.__name__[: -len("Pagination")].lower()
        return None



from unittest.mock import patch

import pytest
from anymail.inbound import AnymailInboundMessage
from anymail.signals import AnymailInboundEvent, EventType, inbound
from django.utils import timezone
from faker import Faker

from ..models import Source
from .test_models import get_byte_stream

fake = Faker()


def conn(sftpsrv):
    """return a dictionary holding argument info for the sftp client"""
    return {
        "host": sftpsrv.host,
        "port": sftpsrv.port,
        "username": "user",
        "password": "pw",
        "default_path": "/home/test",
    }


# filesystem served by pytest-sftpserver plugin
VFS = {
    "home": {
        "test": {
            "pub": {
                "make.txt": "content of make.txt",
                "foo1.csv": "CSV Content",
                "image01.jpg": "data for image01.jpg",
            },
            "read.me": "contents of read.me",
        }
    }
}


@pytest.mark.django_db
class TestBackend:
    def test_mail_backend(self):
        from wbcore.contrib.io.import_export.backends.mail import DataBackend as MailDataBackend

        b1 = get_byte_stream()
        b2 = get_byte_stream()
        filename1 = "file_valid_12385.json"
        filename2 = "file.json"
        att1 = AnymailInboundMessage.construct_attachment(
            "application/json", b1.getvalue(), charset="utf-8", filename=filename1
        )

        att2 = AnymailInboundMessage.construct_attachment(
            "application/json", b2.getvalue(), charset="utf-8", filename=filename2
        )

        msg = AnymailInboundMessage.construct(
            from_email="from@example.com",
            to="to@example.com",
            cc="cc@example.com",
            subject="test subject",
            attachments=[att1, att2],
        )
        backend = MailDataBackend()
        res = list(backend.get_files(timezone.now(), file_name_regex="file_valid_([A-Z0-9]*).json", message=msg))
        assert len(res) == 1
        assert res[0][0] == filename1
        assert (res[0][1]).getvalue() == b1.getvalue()

    @pytest.mark.parametrize(
        "source__import_parameters,from_email",
        [
            ({"whitelisted_emails": ["test@test.ch"]}, "test@test.ch"),
            ({"whitelisted_emails": ["test@test.ch"]}, "spam@    test.ch"),
        ],
    )
    @patch.object(Source, "trigger_workflow")
    def test_inbound_mail(self, mock_process_source, source, from_email):
        from wbcore.contrib.io.models import DataBackend

        data_backend = DataBackend.objects.get(title="Default Mail")

        assert data_backend.backend_class_path == "wbcore.contrib.io.import_export.backends.mail"

        source.data_backend = data_backend
        source.save()
        message = AnymailInboundMessage.construct(
            from_email=from_email,
            to="to@example.com",
            subject=fake.paragraph(),
            text=fake.paragraph(),
            html=f"<p>{fake.paragraph()}</p>",
        )
        event = AnymailInboundEvent(
            event_type=EventType.INBOUND,
            timestamp=None,
            event_id=fake.uuid4(),
            esp_event={"from_email": from_email},
            message=message,
        )
        inbound.send(sender="TEST", event=event, esp_name="test")
        if source.import_parameters.get("whitelisted_emails", [None])[0] == from_email:
            assert mock_process_source.call_count == 1
        else:
            assert mock_process_source.call_count == 0

    def test_sftp_backend_without_credential(self):
        from wbcore.contrib.io.import_export.backends.sftp import DataBackend as SFTPDataBackend

        with pytest.raises(ValueError):
            SFTPDataBackend()

    def test_sftp_backend(self, sftpserver, import_credential_factory):
        from wbcore.contrib.io.import_export.backends.sftp import DataBackend as SFTPDataBackend

        with sftpserver.serve_content(VFS):
            import_credential = import_credential_factory.create(
                username="user",
                password="password",
                additional_resources={
                    "host": sftpserver.host,
                    "port": sftpserver.port,
                },
            )
            copied_file_content = VFS["home"]["test"]["pub"]["foo1.csv"]
            backend = SFTPDataBackend(import_credential=import_credential)
            res = list(
                backend.get_files(timezone.now(), "/home/test/pub", file_name_regex=r".+\.csv$", cleanup_files=True)
            )

            assert len(res) == 1
            assert res[0][0] == "foo1.csv"
            assert (res[0][1]).getvalue().decode("utf-8") == copied_file_content



from .filters import add_filters
from .instance_buttons import add_button, add_instance_button
from .models import get_dependant_dynamic_fields_instances, pre_collection
from .permissions import load_permission_objects
from .serializers import add_additional_resource, add_instance_additional_resource, add_dynamic_button
from .merge import pre_merge
from .clone import post_clone



from configurations import values
from django.db.models import options


class NotificationConfiguration:
    options.DEFAULT_NAMES = options.DEFAULT_NAMES + ("notification_types",)
    NOTIFICATION_BACKEND = values.Value(
        "wbcore.contrib.notifications.backends.firebase.NotificationBackend", environ_prefix=None
    )



from configurations import Configuration, values

from wbcore.configurations import configurations
from wbcore.contrib.agenda.configurations import AgendaConfigurationMixin
from wbcore.contrib.authentication.configurations import (
    AuthenticationConfigurationMixin,
)
from wbcore.contrib.directory.configurations import DirectoryConfigurationMixin
from wbcore.contrib.guardian.configurations import Guardian
from wbcore.contrib.io.configurations import ImportExportBaseConfiguration
from wbcore.contrib.notifications.configurations import NotificationConfiguration


class DevBaseConfiguration(
    AgendaConfigurationMixin,
    AuthenticationConfigurationMixin,
    ImportExportBaseConfiguration,
    DirectoryConfigurationMixin,
    NotificationConfiguration,
    configurations.Base,
    configurations.DevApps,
    configurations.S3Media,
    configurations.Network,
    configurations.Restframework,
    configurations.ConsoleEmail,
    configurations.Templates,
    configurations.Authentication,
    configurations.DevMiddleware,
    configurations.LocalStaticfiles,
    Guardian,
    configurations.WBCore,
    configurations.Celery,
    configurations.I18NL10N,
    configurations.Maintenance,
    configurations.Cache,
    Configuration,
):
    DEFAULT_AUTO_FIELD = "django.db.models.AutoField"
    DEV_USER = values.Value(None, environ_prefix=None)
    ADD_REVERSION_ADMIN = True


class ProductionBaseConfiguration(
    configurations.Uvicorn, configurations.SSLNetwork, configurations.S3Staticfiles, DevBaseConfiguration
):
    pass



from wbcore import serializers as wb_serializers

from ..models import User, UserActivity
from .users import UserRepresentationSerializer


class UserActivitySerializer(wb_serializers.ModelSerializer):
    IP = wb_serializers.CharField(read_only=True)
    _user = UserRepresentationSerializer(source="user")
    user = wb_serializers.PrimaryKeyRelatedField(queryset=User.objects.all(), many=False)
    time_online_minute = wb_serializers.CharField(read_only=True)

    class Meta:
        model = UserActivity
        fields = (
            "id",
            "IP",
            "date",
            "user",
            "_user",
            "status",
            "type",
            "user_agent_info",
            "latest_refresh",
            "time_online_minute",
        )


class UserActivityTableSerializer(UserActivitySerializer):
    today_activity = wb_serializers.CharField(read_only=True)
    beforeyesterday_activity = wb_serializers.CharField(read_only=True)
    yesterday_activity = wb_serializers.CharField(read_only=True)
    user_repr = wb_serializers.CharField(read_only=True)

    class Meta:
        model = User
        fields = ("id", "user_repr", "today_activity", "yesterday_activity", "beforeyesterday_activity")



from .bases import ButtonConfig
from .buttons import ActionButton, DropDownButton, HyperlinkButton, WidgetButton
from .enums import ButtonDefaultColor, ButtonType, HyperlinkTarget
from .view_config import ButtonViewConfig



from unittest.mock import patch

import pytest
from wbcore.contrib.authentication.factories import GroupFactory, UserFactory
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.contrib.workflow.factories import UserStepFactory
from wbcore.contrib.workflow.models import Process, ProcessStep


@pytest.mark.django_db
class TestProcess:
    @patch("wbcore.contrib.workflow.models.process.ProcessStep.get_all_assignees")
    @patch("wbcore.contrib.workflow.models.step.UserStep.notify_assignees")
    def test_set_failed(
        self,
        mock_notify,
        mock_assignees,
        process_step_factory,
        process_factory,
        random_child_step_factory,
        user_step_factory,
    ):
        step = random_child_step_factory(exclude_factories=[UserStepFactory])
        user_step_no_notify = user_step_factory(notify_user=False)
        user_step_notify = user_step_factory(notify_user=True)
        process = process_factory(state=Process.ProcessState.ACTIVE)
        process_step1 = process_step_factory(
            state=ProcessStep.StepState.ACTIVE, process=process, step=step, error_message=None
        )
        process_step2 = process_step_factory(
            state=ProcessStep.StepState.WAITING, process=process, step=step, error_message=None
        )
        process_step3 = process_step_factory(
            state=ProcessStep.StepState.CANCELED, process=process, step=step, error_message=None
        )
        process_step4 = process_step_factory(
            state=ProcessStep.StepState.FINISHED, process=process, step=step, error_message=None
        )
        process_step5 = process_step_factory(state=ProcessStep.StepState.ACTIVE, step=step, error_message=None)
        process_step6 = process_step_factory(
            state=ProcessStep.StepState.ACTIVE, process=process, step=user_step_no_notify, error_message=None
        )
        process_step7 = process_step_factory(
            state=ProcessStep.StepState.ACTIVE, process=process, step=user_step_notify, error_message=None
        )
        user = UserFactory()
        mock_assignees.return_value = [user]
        process.set_failed()
        assert mock_notify.call_count == 1
        assert mock_notify.call_args.args[0] == [user]
        assert str(mock_notify.call_args.args[1].pk) == process_step7.pk
        assert mock_assignees.call_count == 1
        for process_step in ProcessStep.objects.filter(
            id__in=[process_step1.pk, process_step2.pk, process_step6.pk, process_step7.pk]
        ):
            assert process_step.state == ProcessStep.StepState.FAILED
            assert process_step.error_message
        for process_step in ProcessStep.objects.filter(id__in=[process_step3.pk, process_step4.pk, process_step5.pk]):
            assert process_step.state != ProcessStep.StepState.FAILED
            assert process_step.error_message is None
        assert Process.objects.get(id=process.pk).state == Process.ProcessState.FAILED


@pytest.mark.django_db
class TestProcessStep:
    def test_get_instance(self, process_step_factory):
        attached_instance = PersonFactory()
        process_step = process_step_factory(process__instance=attached_instance)
        assert process_step.get_instance() == attached_instance

    def test_get_instance_no_attached(self, process_step_factory):
        process_step = process_step_factory(process__instance=None, process__instance_id=None)
        assert process_step.get_instance() is None

    def test_get_all_assignees_assignee(self, process_step_factory):
        user = UserFactory()
        group = GroupFactory()
        process_step = process_step_factory(assignee=user, group=group)
        assert process_step.get_all_assignees() == [user]

    def test_get_all_assignees_group(self, process_step_factory):
        group = GroupFactory()
        user1 = UserFactory(groups=[group])
        user2 = UserFactory(groups=[group])
        UserFactory()
        process_step = process_step_factory(group=group, assignee=None)
        assert set(process_step.get_all_assignees()) == {user1, user2}

    def test_get_all_assignees_empty_group(self, process_step_factory):
        group = GroupFactory()
        process_step = process_step_factory(group=group, assignee=None)
        assert process_step.get_all_assignees() == []

    def test_get_all_assignees_none(self, process_step_factory):
        group = GroupFactory()
        UserFactory(groups=[group])
        process_step = process_step_factory(group=None, assignee=None)
        assert process_step.get_all_assignees() == []



from django.conf import settings
from django.db import DEFAULT_DB_ALIAS, router
from django.contrib.contenttypes.management import create_contenttypes
from django.apps import apps as global_apps


def _get_all_permissions(opts):
    if action := getattr(settings, "IMPORT_EXPORT_IMPORT_PERMISSION_CODE", None):
        yield "%s_%s" % (action, opts.model_name), "Import %s %s" % (action, opts.verbose_name_raw),
    if action := getattr(settings, "IMPORT_EXPORT_EXPORT_PERMISSION_CODE", None):
        yield "%s_%s" % (action, opts.model_name), "Export %s %s" % (action, opts.verbose_name_raw),

    yield "select_%s" % (opts.model_name,), "Can select %s" % (opts.verbose_name_raw,)


def create_wbcore_permissions(
    app_config,
    verbosity=2,
    interactive=True,
    using=DEFAULT_DB_ALIAS,
    apps=global_apps,
    **kwargs,
):
    if not app_config.models_module:
        return

    # Ensure that contenttypes are created for this app. Needed if
    # 'django.contrib.auth' is in INSTALLED_APPS before
    # 'django.contrib.contenttypes'.
    create_contenttypes(
        app_config,
        verbosity=verbosity,
        interactive=interactive,
        using=using,
        apps=apps,
        **kwargs,
    )

    app_label = app_config.label
    try:
        app_config = apps.get_app_config(app_label)
        ContentType = apps.get_model("contenttypes", "ContentType")
        Permission = apps.get_model("auth", "Permission")
    except LookupError:
        return

    if not router.allow_migrate_model(using, Permission):
        return

    # This will hold the permissions we're looking for as
    # (content_type, (codename, name))
    searched_perms = []
    # The codenames and ctypes that should exist.
    ctypes = set()
    for klass in app_config.get_models():
        # Force looking up the content types in the current database
        # before creating foreign keys to them.
        ctype = ContentType.objects.db_manager(using).get_for_model(klass, for_concrete_model=False)

        ctypes.add(ctype)
        for perm in _get_all_permissions(klass._meta):
            searched_perms.append((ctype, perm))

    # Find all the Permissions that have a content_type for a model we're
    # looking for.  We don't need to check for codenames since we already have
    # a list of the ones we're going to create.
    all_perms = set(
        Permission.objects.using(using)
        .filter(
            content_type__in=ctypes,
        )
        .values_list("content_type", "codename")
    )

    perms = []
    for ct, (codename, name) in searched_perms:
        if (ct.pk, codename) not in all_perms:
            permission = Permission()
            permission._state.db = using
            permission.codename = codename
            permission.name = name
            permission.content_type = ct
            perms.append(permission)

    Permission.objects.using(using).bulk_create(perms)
    if verbosity >= 2:
        for perm in perms:
            print("Adding permission '%s'" % perm)



import factory
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django_celery_beat.models import CrontabSchedule
from faker import Faker

from wbcore.contrib.io.imports import ImportExportHandler
from wbcore.contrib.io.mixins import ImportMixin

from .enums import ExportFormat
from .models import (
    DataBackend,
    ExportSource,
    ImportCredential,
    ImportSource,
    ParserHandler,
    Provider,
    Source,
)


class ImportExportModelHandler(ImportExportHandler):
    MODEL_APP_LABEL = "io.ImportModel"

    def _deserialize(self, data):
        data["relationship"] = ParserHandler.objects.get(id=data["relationship"])
        data["many_relationships"] = [ParserHandler.objects.get(id=d) for d in data["many_relationships"]]

    def _get_instance(self, data, history=None, **kwargs) -> models.Model | None:
        if id := data.get("id", None):
            if history:
                return history.filter(id=id).first()
            return self.model.objects.filter(id=id).first()

    def _create_instance(self, data, **kwargs) -> models.Model:
        data.pop("id", None)
        many_relationships = data.pop("many_relationships", [])
        obj = self.model.objects.create(import_source=self.import_source, **data)
        for rel in many_relationships:
            obj.many_relationships.add(rel)
        return obj

    def _get_history(self, history):
        # Return all object with the same relationship
        return self.model.objects.filter(relationship=history["relationship"])


# TODO remove this class
class ImportModel(ImportMixin):
    import_export_handler_class = ImportExportModelHandler
    relationship = models.ForeignKey(
        "io.ParserHandler", blank=True, null=True, on_delete=models.SET_NULL, related_name="imports"
    )
    import_source = models.ForeignKey(
        "io.ImportSource", blank=True, null=True, on_delete=models.SET_NULL, related_name="imports"
    )
    many_relationships = models.ManyToManyField("io.ParserHandler", blank=True, related_name="many_imports")
    number = models.FloatField(blank=True, null=True)
    text = models.TextField(blank=True, null=True)
    name = models.CharField(max_length=256, blank=True, null=True)


fake = Faker()


class CrontabScheduleFactory(factory.django.DjangoModelFactory):
    hour = factory.Iterator(range(0, 24))
    minute = factory.Iterator(range(0, 60))

    class Meta:
        model = CrontabSchedule


class ProviderFactory(factory.django.DjangoModelFactory):
    title = factory.Faker("company")
    key = factory.LazyAttribute(lambda x: x.title.lower())

    class Meta:
        model = Provider


class DataBackendFactory(factory.django.DjangoModelFactory):
    title = factory.Sequence(lambda n: f"Data Backend {n}")
    save_data_in_import_source = False
    passive_only = False
    backend_class_path = "wbcore.contrib.io.backends.abstract"
    backend_class_name = "AbstractDataBackend"
    provider = factory.SubFactory("wbcore.contrib.io.factories.ProviderFactory")

    class Meta:
        model = DataBackend


class ExportSourceFactory(factory.django.DjangoModelFactory):
    """
    Default exportt source factory that are attached to the parser handler model and serializer
    """

    created = factory.Faker("date_time")
    last_updated = factory.Faker("date_time")
    log = factory.Faker("paragraph")
    origin = factory.Faker("company")
    status = ExportSource.Status.PENDING.value
    creator = factory.SubFactory("wbcore.contrib.authentication.factories.UserFactory")
    file = None
    format = factory.Iterator(ExportFormat.values)

    content_type = factory.LazyAttribute(lambda x: ContentType.objects.get_for_model(ParserHandler))
    resource_path = "wbcore.contrib.io.resources.ViewResource"
    resource_kwargs = {
        "columns_map": {"parser": "Parser", "handler": "Handler"},
        "serializer_class_path": "wbcore.contrib.io.serializers.ParserHandlerModelSerializer",
    }
    query_str = factory.LazyAttribute(lambda x: ParserHandler.objects.all().query.sql_with_params()[0])
    query_params = factory.LazyAttribute(lambda x: ParserHandler.objects.all().query.sql_with_params()[1])

    class Meta:
        model = ExportSource


class ImportSourceFactory(factory.django.DjangoModelFactory):
    created = factory.Faker("date_time")
    last_updated = factory.Faker("date_time")
    parser_handler = factory.SubFactory("wbcore.contrib.io.factories.ParserHandlerFactory")
    source = factory.SubFactory("wbcore.contrib.io.factories.SourceFactory")
    save_data = True
    log = factory.Faker("paragraph")
    origin = factory.Faker("company")
    file = None

    class Meta:
        model = ImportSource


class ImportCredentialFactory(factory.django.DjangoModelFactory):
    type = ImportCredential.Type.CREDENTIAL
    username = factory.Sequence(lambda n: f"username-{n}")
    password = factory.Sequence(lambda n: f"password-{n}")
    key = factory.Sequence(lambda n: f"credential-{n}")
    validity_start = None
    validity_end = None

    class Meta:
        model = ImportCredential


class SourceFactory(factory.django.DjangoModelFactory):
    crontab = factory.SubFactory("wbcore.contrib.io.factories.CrontabScheduleFactory")
    data_backend = factory.SubFactory("wbcore.contrib.io.factories.DataBackendFactory")
    import_parameters = factory.LazyAttribute(lambda o: {fake.name(): fake.name(), fake.name(): fake.name()})
    connection_parameters = factory.LazyAttribute(lambda o: {fake.name(): fake.name(), fake.name(): fake.name()})
    is_active = True
    title = factory.Sequence(lambda n: f"title {n}")

    class Meta:
        model = Source

    @factory.post_generation
    def parser_handler(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for parser_handler in extracted:
                self.parser_handler.add(parser_handler)

    @factory.post_generation
    def credentials(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for credential in extracted:
                self.credentials.add(credential)


class ParserHandlerFactory(factory.django.DjangoModelFactory):
    parser = factory.Sequence(lambda n: f"parser_{n}")
    handler = "io.ImportModel"
    allow_file_type = None

    class Meta:
        model = ParserHandler


class ImportModelFactory(factory.django.DjangoModelFactory):
    relationship = factory.SubFactory("wbcore.contrib.io.factories.ParserHandlerFactory")
    import_source = factory.SubFactory("wbcore.contrib.io.factories.ImportSourceFactory")
    number = factory.Faker("pyfloat")
    text = factory.Faker("paragraph")
    name = factory.Faker("name")

    class Meta:
        model = ImportModel

    @factory.post_generation
    def many_relationships(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for relationship in extracted:
                self.many_relationships.add(relationship)



from typing import Callable

from django.db.models.signals import ModelSignal


class SignalWrapper:
    """A convenience class that encapsulates a signal in order to
    be used in the DisconnectedSignals context manager
    """

    def __init__(self, signal: ModelSignal, receiver: Callable, sender, dispatch_uid: str | None = None):
        self.signal = signal
        self.receiver = receiver
        self.sender = sender
        self.dispatch_uid = dispatch_uid

    def connect(self):
        self.signal.connect(receiver=self.receiver, sender=self.sender, weak=False, dispatch_uid=self.dispatch_uid)

    def disconnect(self):
        self.signal.disconnect(receiver=self.receiver, sender=self.sender, dispatch_uid=self.dispatch_uid)


class DisconnectSignals:
    """A Context-Manager to temporarily disable a set of signals"""

    def __init__(self, signals: list[SignalWrapper]):
        self.signals = signals

    def __enter__(self):
        for signal in self.signals:
            signal.disconnect()

    def __exit__(self, type, value, traceback):
        for signal in self.signals:
            signal.connect()



from django.utils.translation import gettext_lazy as _

from wbcore import filters
from wbcore.contrib.example_app.models import Stadium, Team
from wbcore.contrib.geography.models import Geography


class StadiumFilter(filters.FilterSet):
    capacity__gte = filters.NumberFilter(label=_("Capacity"), lookup_expr="gte", field_name="capacity")
    capacity__lte = filters.NumberFilter(label=_("Capacity"), lookup_expr="lte", field_name="capacity")
    teams_playing = filters.ModelChoiceFilter(
        label=_("Team Playing"),
        queryset=Team.objects.all(),
        endpoint=Team.get_representation_endpoint(),
        value_key=Team.get_representation_value_key(),
        label_key=Team.get_representation_label_key(),
    )
    city = filters.ModelMultipleChoiceFilter(
        label=_("Cities"),
        queryset=Geography.cities.all(),
        endpoint=Geography.get_representation_endpoint(),
        value_key=Geography.get_representation_value_key(),
        label_key=Geography.get_representation_label_key(),
        filter_params={"level": 3},
    )

    class Meta:
        model = Stadium
        fields = {
            "name": ["exact", "icontains"],
        }



from django.contrib.auth.models import Group
from django.db.models.query import QuerySet

from wbcore.contrib.authentication.models.users import User


def get_internal_groups() -> QuerySet[Group]:
    from wbcore.permissions.registry import user_registry

    """
    Return the cached groups of internals users

    Returns:
        A queryset of group corresponding to the internal notion defined by the set UserBackend

    Raises:
        ValueError: If user backend path does not correspond to a valid module
    """
    return user_registry.internal_groups


def get_internal_users() -> QuerySet[User]:
    from wbcore.permissions.registry import user_registry

    """
    Return the cached queryset of internals users

    Returns:
        A queryset of user corresponding to the internal notion defined by the set UserBackend

    Raises:
        ValueError: If user backend path does not correspond to a valid module
    """
    return user_registry.internal_users


def is_internal_user(user: User, include_superuser: bool = False) -> bool:
    return user and (user.is_internal or (include_superuser and user.is_superuser))



from datetime import date, timedelta

from django.contrib import admin

from wbcore.contrib.notifications.tasks import send_notification_task

from .models import (
    Notification,
    NotificationType,
    NotificationTypeSetting,
    NotificationUserToken,
)


@admin.register(Notification)
class NotificationModelAdmin(admin.ModelAdmin):
    search_fields = ["title", "body"]

    list_display = [
        "title",
        "user",
        "notification_type",
        "endpoint",
        "sent",
        "read",
    ]

    @admin.action(description="Send Notification")
    def send_notification(self, request, queryset):
        for notification in queryset:
            send_notification_task.delay(notification.id)

    actions = [send_notification]


@admin.register(NotificationUserToken)
class NotificationUserTokenModelAdmin(admin.ModelAdmin):
    list_display = ("user", "token", "device_type", "updated")

    @admin.action(description="Remove all stale tokens")
    def remove_stale_tokens(self, request, queryset):
        cutoff_date = date.today() - timedelta(days=60)
        queryset.filter(updated__date__lte=cutoff_date).delete()

    actions = [remove_stale_tokens]


@admin.register(NotificationTypeSetting)
class NotificationTypeSettingModelAdmin(admin.ModelAdmin):
    list_display = ("notification_type", "user", "enable_web", "enable_mobile", "enable_email")


@admin.register(NotificationType)
class NotificationTypeModelAdmin(admin.ModelAdmin):
    list_display = (
        "code",
        "title",
        "contenttype",
        "default_enable_web",
        "default_enable_mobile",
        "default_enable_email",
        "is_lock",
    )



from contextlib import suppress

from django.contrib.auth import get_user_model
from django.db import models
from django.db.models import UniqueConstraint
from django.db.models.signals import post_save
from django.dispatch import receiver
from guardian.utils import get_anonymous_user

from wbcore.models import WBModel


class NotificationType(WBModel):
    code = models.CharField(max_length=128, unique=True)
    title = models.CharField(max_length=128, default="")
    help_text = models.CharField(max_length=512, default="")

    stale = models.BooleanField(default=False)
    contenttype = models.ForeignKey(
        to="contenttypes.ContentType", related_name="+", null=True, blank=True, on_delete=models.SET_NULL
    )
    resource_button_label = models.CharField(max_length=128)

    default_enable_web = models.BooleanField(default=False)
    default_enable_mobile = models.BooleanField(default=False)
    default_enable_email = models.BooleanField(default=False)
    is_lock = models.BooleanField(default=False)

    def __str__(self) -> str:
        return f"{self.title}"

    def save(self, *args, **kwargs):
        if not self.resource_button_label:
            if self.contenttype:
                self.resource_button_label = f"Open {self.contenttype.name}"
            else:
                self.resource_button_label = "Open Resource"
        super().save(*args, **kwargs)

    class Meta:
        verbose_name = "Notification Type"
        verbose_name_plural = "Notification Types"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:notifications:notification_type"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:notifications:notification_type_representation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{title}}"


class NotificationTypeSetting(WBModel):
    notification_type = models.ForeignKey(
        to="notifications.NotificationType", related_name="user_settings", on_delete=models.CASCADE
    )
    user = models.ForeignKey(
        to=get_user_model(), related_name="wbnotification_user_settings", on_delete=models.CASCADE
    )

    enable_web = models.BooleanField(default=False)
    enable_mobile = models.BooleanField(default=False)
    enable_email = models.BooleanField(default=False)

    def __str__(self) -> str:
        return f"{self.user}: {self.notification_type} ({self.enable_web}/{self.enable_mobile}/{self.enable_email})"

    class Meta:
        verbose_name = "Notification Type Setting"
        verbose_name_plural = "Notification Type Settings"
        constraints = [
            UniqueConstraint(
                name="unique_notification_setting",
                fields=["user", "notification_type"],
            )
        ]

    def save(self, *args, **kwargs):
        if self.notification_type.is_lock:
            self.enable_web = self.notification_type.default_enable_web
            self.enable_mobile = self.notification_type.default_enable_mobile
            self.enable_email = self.notification_type.default_enable_email

        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:notifications:notification_type_setting"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:notifications:notification_type_setting_representation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{notification_type}}"


@receiver(post_save, sender="notifications.NotificationType")
def post_save_notification_type(instance, **kwargs):
    anonymous_user = get_anonymous_user()

    objs = []
    for user in get_user_model().objects.filter(~models.Q(pk=anonymous_user.pk)):
        try:
            existing_setting = NotificationTypeSetting.objects.get(user=user, notification_type=instance)
            enable_web = existing_setting.enable_web if not instance.is_lock else instance.default_enable_web
            enable_mobile = existing_setting.enable_mobile if not instance.is_lock else instance.default_enable_mobile
            enable_email = existing_setting.enable_email if not instance.is_lock else instance.default_enable_email
        except NotificationTypeSetting.DoesNotExist:
            enable_web = instance.default_enable_web
            enable_mobile = instance.default_enable_mobile
            enable_email = instance.default_enable_email

        objs.append(
            NotificationTypeSetting(
                notification_type=instance,
                user=user,
                enable_web=enable_web,
                enable_mobile=enable_mobile,
                enable_email=enable_email,
            )
        )
    if objs:
        NotificationTypeSetting.objects.bulk_create(
            objs,
            unique_fields=["notification_type", "user"],
            update_fields=["enable_web", "enable_mobile", "enable_email"],
            update_conflicts=True,
        )


@receiver(post_save, sender=get_user_model())
def post_save_user(sender, instance, created, raw, **kwargs):
    if created:
        with suppress(get_user_model().DoesNotExist):
            anonymous_user = get_anonymous_user()
            if instance.pk == anonymous_user.pk:
                return

        objs = []
        for notification_type in NotificationType.objects.all():
            objs.append(
                NotificationTypeSetting(
                    notification_type=notification_type,
                    user=instance,
                    enable_web=notification_type.default_enable_web,
                    enable_mobile=notification_type.default_enable_mobile,
                    enable_email=notification_type.default_enable_email,
                )
            )
        if objs:
            NotificationTypeSetting.objects.bulk_create(objs, unique_fields=["notification_type", "user"])



from .display import Display
from .layouts import Inline, Layout, Section
from .operators import default
from .pages import Page
from .styles import Style


def create_simple_display(
    grid_template_areas: list[list[str]],
    sections: list[Section] | None = None,
    inlines: list[Inline] | None = None,
) -> Display:
    """Creates a simple display without having to specify all the intermediary classes

    Args:
        grid_template_areas: 2D string array containing all fields that should be displayed
        sections: A list of sections included in the `grid_template_areas`
        inlines: A list of inlines if necessary.

    Returns:
        A `Display` containing 1 `Page` with all fields in 1 `Section` with all columns and rows having the same dimensions

    """
    return Display(
        pages=[
            Page(
                layouts={
                    default(): Layout(
                        grid_template_areas=grid_template_areas,
                        grid_auto_columns=Style.fr(1),
                        inlines=inlines if inlines else [],
                        sections=sections if sections else [],
                    )
                }
            )
        ]
    )


def create_simple_section(
    key: str,
    title: str,
    grid_template_areas: list[list[str]] | None = None,
    inline_key: str | None = None,
    extra_display_kwargs: dict | None = None,
    **kwargs,
) -> Section:
    """Creates a simple section without having to specify everything

    Args:
        key: A string that will be later referenced in the display's `grid_template_areas`
        title: A string which is used as a header
        grid_template_areas: 2D string array containing all fields that should be displayed
        inline_key: if the section is a nested table, the inline is the corresponding endpoint
        extra_display_kwargs: an additional dictionary for create_simple_display parameters.

    Returns:
        A `Section` containing a simple `Display`

    """
    return Section(
        key=key,
        title=title,
        display=create_simple_display(
            grid_template_areas=grid_template_areas if grid_template_areas else [],
            inlines=[Inline(key=inline_key, endpoint=inline_key)] if inline_key else [],
            **extra_display_kwargs if extra_display_kwargs else {},
        ),
        **kwargs,
    )


def create_simple_page_with_inline(title: str, key: str) -> Page:
    return Page(
        title=title,
        layouts={
            default(): Layout(
                grid_template_areas=[[key]], grid_template_rows=["100%"], inlines=[Inline(key=key, endpoint=key)]
            )
        },
    )



# Generated by Django 5.0.8 on 2024-09-10 10:44

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("contenttypes", "0002_remove_content_type_name"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            database_operations=[],
            state_operations=[
                migrations.CreateModel(
                    name="GroupObjectPermission",
                    fields=[
                        ("object_pk", models.CharField(max_length=255, verbose_name="object ID")),
                        ("id", models.BigAutoField(editable=False, primary_key=True, serialize=False, unique=True)),
                        ("editable", models.BooleanField(default=True)),
                        (
                            "content_type",
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.CASCADE, to="contenttypes.contenttype"
                            ),
                        ),
                        ("group", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="auth.group")),
                        (
                            "permission",
                            models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="auth.permission"),
                        ),
                    ],
                    options={
                        "db_table": "bridger_biggroupobjectpermission",
                        "abstract": False,
                        "default_related_name": "groupobjectpermissions",
                        "unique_together": {("group", "permission", "object_pk")},
                    },
                ),
                migrations.CreateModel(
                    name="UserObjectPermission",
                    fields=[
                        ("object_pk", models.CharField(max_length=255, verbose_name="object ID")),
                        ("id", models.BigAutoField(editable=False, primary_key=True, serialize=False, unique=True)),
                        ("editable", models.BooleanField(default=True)),
                        ("system", models.BooleanField(default=False)),
                        (
                            "content_type",
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.CASCADE, to="contenttypes.contenttype"
                            ),
                        ),
                        (
                            "permission",
                            models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="auth.permission"),
                        ),
                        (
                            "user",
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL
                            ),
                        ),
                    ],
                    options={
                        "db_table": "bridger_biguserobjectpermission",
                        "abstract": False,
                        "default_related_name": "userobjectpermissions",
                        "unique_together": {("user", "permission", "object_pk")},
                    },
                ),
            ],
        ),
        migrations.AlterModelTable(
            name="groupobjectpermission",
            table=None,
        ),
        migrations.AlterModelTable(
            name="userobjectpermission",
            table=None,
        ),
        migrations.AddIndex(
            model_name="groupobjectpermission",
            index=models.Index(fields=["content_type", "object_pk"], name="wbcore_guar_content_4afd41_idx"),
        ),
        migrations.AddIndex(
            model_name="groupobjectpermission",
            index=models.Index(fields=["content_type", "object_pk", "group"], name="wbcore_guar_content_83dcd4_idx"),
        ),
        migrations.AddIndex(
            model_name="userobjectpermission",
            index=models.Index(fields=["content_type", "object_pk"], name="wbcore_guar_content_138d49_idx"),
        ),
        migrations.AddIndex(
            model_name="userobjectpermission",
            index=models.Index(fields=["content_type", "object_pk", "user"], name="wbcore_guar_content_98fc0f_idx"),
        ),
    ]



import json
import os

import firebase_admin
from django.utils.html import strip_tags
from firebase_admin import messaging
from firebase_admin.credentials import Certificate
from firebase_admin.exceptions import InvalidArgumentError

from wbcore.contrib.notifications.backends.abstract_backend import (
    AbstractNotificationBackend,
)
from wbcore.contrib.notifications.models import Notification, NotificationUserToken
from wbcore.contrib.notifications.models.notification_types import (
    NotificationTypeSetting,
)


class NotificationBackend(AbstractNotificationBackend):
    FCM_OPTIONS = messaging.FCMOptions(analytics_label="notification")

    @classmethod
    def get_firebase_app(cls, certificate: Certificate):
        try:
            return firebase_admin.get_app()
        except ValueError:
            return firebase_admin.initialize_app(certificate)

    @classmethod
    def get_firebase_credentials(cls):
        certificate = json.loads(os.environ.get("FIREBASE_SERVICE_WORKER_CREDENTIALS", "{}"))
        return Certificate(certificate)

    @classmethod
    def send_notification(cls, notification: Notification):
        app = cls.get_firebase_app(cls.get_firebase_credentials())
        notification_user_settings = NotificationTypeSetting.objects.get(
            notification_type=notification.notification_type,
            user=notification.user,
        )
        tokens = NotificationUserToken.objects.filter_for_user_settings(notification_user_settings)
        endpoint_data = {}  # Firebase can't accept non-string value
        if full_endpoint := notification.get_full_endpoint():
            endpoint_data["endpoint"] = full_endpoint
        expired_tokens = []
        for token in tokens.filter(device_type=NotificationUserToken.NotificationDeviceType.MOBILE):
            message = messaging.Message(
                notification=messaging.Notification(
                    title=notification.title,
                    body=strip_tags(notification.body or ""),
                ),
                data=endpoint_data,
                token=token.token,
                android=messaging.AndroidConfig(
                    priority="high",
                    ttl=3600,
                    notification=messaging.AndroidNotification(
                        color="#f45342",
                        default_sound=True,
                        visibility="public",
                        default_vibrate_timings=True,
                        priority="max",
                    ),
                ),
                apns=messaging.APNSConfig(
                    payload=messaging.APNSPayload(
                        aps=messaging.Aps(sound=messaging.CriticalSound("default", volume=1.0))
                    ),
                ),
                fcm_options=cls.FCM_OPTIONS,
            )
            try:
                messaging.send(message, False, app)
            except (messaging.UnregisteredError, messaging.QuotaExceededError):
                expired_tokens.append(token)
            except InvalidArgumentError:  # this happens if the body is too big for the mobile push
                pass

        for token in tokens.filter(device_type=NotificationUserToken.NotificationDeviceType.WEB):
            data = {
                "title": notification.title,
                "body": strip_tags(notification.body or ""),
                "is_endpoint_internal": (
                    "true" if notification.is_endpoint_internal else "false"
                ),  # we need the data dictionary to contains only string values, otherwise firebase API complains
            }
            data.update(endpoint_data)
            message = messaging.Message(
                data=data,
                token=token.token,
                fcm_options=cls.FCM_OPTIONS,
            )
            try:
                messaging.send(message, False, app)
            except (messaging.UnregisteredError, messaging.QuotaExceededError):
                expired_tokens.append(token)

        for expired_token in expired_tokens:
            expired_token.delete()

    @classmethod
    def get_configuration(cls) -> dict:
        return {
            "firebase_config": json.loads(os.environ.get("FIREBASE_WEB_CONFIG", "")),
            "vapid_key": os.environ.get("FIREBASE_VAPID_KEY", ""),
        }



from typing import TYPE_CHECKING

from django.utils.functional import cached_property
from rest_framework import serializers

if TYPE_CHECKING:
    from wbcore.contrib.authentication.models import User
    from wbcore.contrib.directory.models import Person


class UserProfileMixin(serializers.Serializer):
    @cached_property
    def user(self) -> "User | None":
        """Returns the user associated with the request."""
        if request := self.context.get("request", None):
            return request.user
        return None

    @cached_property
    def profile(self) -> "Person | None":
        """Returns the profile of the user associated with the request."""
        if user := self.user:
            return user.profile
        return None



from django.utils.translation import gettext as _

from wbcore.metadata.configs.titles import TitleViewConfig

from ...models import User


class UserActivityModelTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("User Activity")


class UserActivityUserModelTitleConfig(TitleViewConfig):
    def get_list_title(self):
        user = User.objects.get(id=self.view.kwargs["user_id"])
        return _("User Activity of {user}").format(user=user.profile.computed_str)


class UserActivityTableTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("User Activity Table")


class UserActivityChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("User Activity Chart")


class UserActivityUserChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        user = User.objects.get(id=self.view.kwargs["user_id"])
        return _("User Activity of {user}").format(user=user.profile.computed_str)



from functools import reduce


def ilen(iterable):
    return reduce(lambda sum, element: sum + 1, iterable, 0)


def leaf_subclasses(parent_class):
    if subclasses := parent_class.__subclasses__():
        for sub_class in subclasses:
            yield from leaf_subclasses(sub_class)
    else:
        yield parent_class



from django.forms import fields as form_fields
from dynamic_preferences.types import BooleanPreference as BaseBooleanPreference
from dynamic_preferences.types import ChoicePreference as BaseChoicePreference
from dynamic_preferences.types import DatePreference as BaseDatePreference
from dynamic_preferences.types import DateTimePreference as BaseDateTimePreference
from dynamic_preferences.types import DecimalPreference as BaseDecimalPreference
from dynamic_preferences.types import DurationPreference as BaseDurationPreference
from dynamic_preferences.types import FilePreference as BaseFilePreference
from dynamic_preferences.types import FloatPreference as BaseFloatPreference
from dynamic_preferences.types import IntegerPreference as BaseIntegerPreference
from dynamic_preferences.types import LongStringPreference as BaseLongStringPreference
from dynamic_preferences.types import ModelChoicePreference
from dynamic_preferences.types import MultipleChoicePreference as BaseMultipleChoicePreference
from dynamic_preferences.types import StringPreference as BaseStringPreference
from dynamic_preferences.types import TimePreference as BaseTimePreference

from wbcore.serializers import fields
from wbcore.serializers.fields import LanguageChoiceField

FIELD_MAPPING = {
    form_fields.BooleanField: fields.BooleanField,
    form_fields.IntegerField: fields.IntegerField,
    form_fields.FloatField: fields.FloatField,
    form_fields.DecimalField: fields.DecimalField,
    form_fields.CharField: fields.CharField,
    form_fields.ChoiceField: fields.ChoiceField,
    form_fields.FileField: fields.FileField,
    form_fields.DurationField: fields.DurationField,
    form_fields.DateField: fields.DateField,
    form_fields.DateTimeField: fields.DateTimeField,
    form_fields.TimeField: fields.TimeField,
    form_fields.MultipleChoiceField: fields.MultipleChoiceField,
}


class PreferenceMixin:
    weight: int = 0

    def get_api_additional_data(self):
        return {}

    def get_api_field_data(self):
        field_class = FIELD_MAPPING[self.field_class]
        kwargs = self.get_field_kwargs()
        del kwargs["widget"]
        rep = field_class(**kwargs, default=kwargs.get("initial")).get_representation(None, self.name)[1]
        rep["required"] = True
        return rep


class BooleanPreference(PreferenceMixin, BaseBooleanPreference):
    pass


class IntegerPreference(PreferenceMixin, BaseIntegerPreference):
    pass


class DecimalPreference(PreferenceMixin, BaseDecimalPreference):
    pass


class FloatPreference(PreferenceMixin, BaseFloatPreference):
    pass


class StringPreference(PreferenceMixin, BaseStringPreference):
    pass


class LongStringPreference(PreferenceMixin, BaseLongStringPreference):
    pass


class ChoicePreference(PreferenceMixin, BaseChoicePreference):
    pass


class FilePreference(PreferenceMixin, BaseFilePreference):
    pass


class DurationPreference(PreferenceMixin, BaseDurationPreference):
    pass


class DatePreference(PreferenceMixin, BaseDatePreference):
    pass


class DateTimePreference(PreferenceMixin, BaseDateTimePreference):
    pass


class TimePreference(PreferenceMixin, BaseTimePreference):
    pass


class MultipleChoicePreference(PreferenceMixin, BaseMultipleChoicePreference):
    pass


class LanguageChoicePreference(PreferenceMixin, BaseChoicePreference):
    def get_api_field_data(self):
        kwargs = self.get_field_kwargs()
        del kwargs["widget"]
        rep = LanguageChoiceField(**kwargs, default=kwargs.get("initial")).get_representation(None, self.name)[1]
        rep["required"] = True
        return rep


class CallableDefaultModelChoicePreference(ModelChoicePreference):
    """
    Propose a ModelChoicePreference class where default is allowed to be a callable property. This avoids unnecessary db calls at start

    This type expects a mandatory model attribute and a section of type dynamic_preferences.Section.
    """

    def __init__(self, registry=None):
        self.registry = registry
        self.queryset = self.model.objects.all()
        self.serializer = self.serializer_class(self.model)
        self._setup_signals()

    @property
    def default(self):
        raise NotImplementedError("Default property needs to be defined for this preference type")



from contextlib import suppress

import tablib
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.dispatch import receiver
from django.http import HttpResponse
from django.utils.functional import cached_property
from django.utils.translation import gettext as _
from import_export.admin import ImportExportMixin
from rest_framework import status
from rest_framework.authentication import SessionAuthentication
from rest_framework.decorators import action, parser_classes
from rest_framework.exceptions import ParseError
from rest_framework.parsers import MultiPartParser
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from wbcore import serializers
from wbcore.contrib.authentication.authentication import JWTCookieAuthentication
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.io.import_export.parsers.base_csv import parse
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.display.instance_display import create_simple_display
from wbcore.signals.instance_buttons import add_extra_button

from .enums import ExportFormat, get_django_import_export_format
from .resources import ViewResource
from .utils import get_import_export_identifier, parse_endpoint


class ImportFileSerializer(serializers.Serializer):
    file = serializers.FileField()


class ExportFileSerializer(serializers.Serializer):
    format = serializers.ChoiceField(choices=ExportFormat.choices, label="Format", required=True)


class ImportExportDRFMixin(ImportExportMixin):
    """
    This mixin can be imported to enable import export out of the box. This will add two action to import a file given its extension
    or Export the filtered queryset content as a file
    """

    IMPORT_ALLOWED: bool = False  # by default import is disabled and requires an explicit defined resource
    EXPORT_ALLOWED: bool = True

    @cached_property
    def model(self):
        try:
            return getattr(self.queryset, "model", None)
        except AttributeError:
            raise ParseError("Malformed Queryset")

    @cached_property
    def opts(self):
        return self.model._meta

    @cached_property
    def export_pagination_limit(self) -> int:
        return getattr(
            settings, "WBIMPORT_EXPORT_DEFAULT_EXPORT_PAGINATION_LIMIT", settings.REST_FRAMEWORK.get("page_size", 25)
        )

    def get_resource_class(self) -> type:
        """
        This function returns return the resource class to use. Default to the model resource factory on the queryset model
        """
        return ViewResource

    def get_import_resource_kwargs(self):
        return {
            "extra_columns_values": self.kwargs,  # contains the extra column to inject into the dataset during parsing
            "columns_mapping": {},  # Store the column mapping to rename dataset column during parsing
        }

    def get_import_parser_path(self) -> str:
        return parse.__module__

    def get_resource_serializer_class(self):
        resource_kwargs = {}
        with suppress(AttributeError, AssertionError):
            # we have to mocky patch the action to be "list" because sometime we differentiate the serializer to use
            previous_action = getattr(self, "action", "list")
            setattr(self, "action", "list")
            serializer_class = getattr(
                self, "serializer_class", self.get_serializer_class()
            )  # we prioritize the default serializer class attribute
            resource_kwargs["serializer_class_path"] = serializer_class.__module__ + "." + serializer_class.__name__
            setattr(self, "action", previous_action)
        return resource_kwargs

    def _get_data_for_export(self, request, queryset, *args, **kwargs) -> tablib.Dataset:
        raise NotImplementedError()

    def has_import_permission(self, request) -> bool:
        """
        called by the import workflow to determine if a user can import data

        Can be overriden if custom behavior is necessary
        """
        return self.IMPORT_ALLOWED and request and super().has_import_permission(request) and "pk" not in self.kwargs

    def has_export_permission(self, request) -> bool:
        """
        called by the import workflow to determine if a user can export data

        Can be overriden if custom behavior is necessary
        """
        return self.EXPORT_ALLOWED and request and super().has_export_permission(request) and "pk" not in self.kwargs

    @action(
        detail=False,
        methods=["PATCH"],
        permission_classes=[IsAuthenticated],
        # authentication_classes=[SessionAuthentication, JWTCookieAuthentication],
        url_name="processimport",
    )
    @parser_classes([MultiPartParser])
    def process_import(self, request, **kwargs):
        """
        This action parse a file and trigger the import workflow from its dataset
        """
        if not self.has_import_permission(request):
            raise PermissionDenied
        if request.data:
            if file_tmp := request.FILES.get("file"):
                from wbcore.contrib.io.models import (  # for circular dependency. Will fix later
                    ImportSource,
                    ParserHandler,
                    import_data_as_task,
                )

                parser_handler, _ = ParserHandler.objects.get_or_create(
                    handler=f"{self.model._meta.app_label}.{self.model.__name__}", parser=self.get_import_parser_path()
                )
                import_source = ImportSource.objects.create(
                    parser_handler=parser_handler,
                    file=file_tmp,
                    origin="Internal Import",
                    creator=request.user,
                    resource_kwargs=self.get_import_resource_kwargs(),
                )
                import_data_as_task.delay(import_source.id)

                return HttpResponse("Your file will be imported shortly", status=status.HTTP_200_OK)
        return HttpResponse("Not all expected parameter were provided", status=status.HTTP_400_BAD_REQUEST)

    @action(
        detail=False,
        methods=["PATCH"],
        permission_classes=[IsAuthenticated],
        authentication_classes=[JWTCookieAuthentication],
        url_name="processexport",
    )
    def process_export(self, request, **kwargs):
        """
        This action export a queryset into a file
        """
        if not self.has_export_permission(request):
            raise PermissionDenied
        if (export_format := int(request.data.get("format", -1))) is not None and export_format in ExportFormat.values:
            queryset = self.filter_queryset(self.get_queryset())
            if queryset.exists():
                from wbcore.contrib.io.models import ExportSource

                # The export job is too big and needs to be ran asynchronously
                try:
                    export_data = self._get_data_for_export(request, queryset, **kwargs)
                    _data = [list(row) for row in export_data._data]

                    ExportSource.objects.create(
                        creator=request.user,
                        format=export_format,
                        content_type=self.get_content_type(),
                        data={"data": _data, "headers": export_data.headers},
                    )
                except NotImplementedError:
                    resource_kwargs = {
                        "columns_map": ViewResource.get_columns_map(self),
                    }
                    resource_kwargs.update(self.get_resource_serializer_class())

                    resource_class = self.get_resource_class()
                    query_str, query_params = queryset.query.sql_with_params()

                    ExportSource.objects.create(
                        creator=request.user,
                        format=export_format,
                        content_type=self.get_content_type(),
                        resource_path=resource_class.__module__ + "." + resource_class.__name__,
                        resource_kwargs=resource_kwargs,
                        query_str=query_str,
                        query_params=query_params,
                    )
                return Response("export job created", status=status.HTTP_200_OK)
        return HttpResponse("Not all expected parameter were provided", status=status.HTTP_400_BAD_REQUEST)

    @action(
        detail=False,
        methods=["GET"],
        permission_classes=[IsAuthenticated],
        authentication_classes=[SessionAuthentication, JWTCookieAuthentication],
        url_name="exporttemplate",
    )
    def export_template(self, request, **kwargs):
        """
        This action export a template into a file
        """
        file_format = get_django_import_export_format(ExportFormat.CSV)()
        export_data = self.get_export_data(file_format, request, None, **kwargs)
        content_type = file_format.get_content_type()
        response = HttpResponse(export_data, content_type=content_type)
        response["Content-Disposition"] = 'attachment; filename="%s"' % (
            self.get_export_filename(request, None, file_format),
        )
        return response


@receiver(add_extra_button)
def add_template_extra_button(sender, instance, request, view, pk=None, **kwargs):
    if view and not instance and hasattr(view, "has_import_permission") and view.has_import_permission(request):
        return bt.HyperlinkButton(
            endpoint=parse_endpoint(request, "export_template", export_format=ExportFormat.CSV),
            title="Import Template (CSV)",
            label="Import Template (CSV)",
        )


@receiver(add_extra_button)
def add_import_extra_button(sender, instance, request, view, pk=None, **kwargs):
    if (
        view
        and not instance
        and hasattr(view, "has_import_permission")
        and view.has_import_permission(request)
        and view.get_resource_class()
    ):
        return bt.ActionButton(
            method=RequestType.PATCH,
            icon=WBIcon.DOWNLOAD.icon,
            identifiers=(get_import_export_identifier(view),),
            endpoint=parse_endpoint(request, "process_import"),
            action_label=_("Import"),
            title=_("Import"),
            label=_("Import"),
            description_fields=_("Please provide a valid CSV file (Coma Separated Value)"),
            serializer=ImportFileSerializer,
            instance_display=create_simple_display([["file"]]),
            weight=0,
        )


@receiver(add_extra_button)
def add_export_extra_button(sender, instance, request, view, pk=None, **kwargs):
    if view and not instance and hasattr(view, "has_export_permission") and view.has_export_permission(request):
        return bt.ActionButton(
            method=RequestType.PATCH,
            icon=WBIcon.DOWNLOAD.icon,
            identifiers=(get_import_export_identifier(view),),
            endpoint=parse_endpoint(request, "process_export"),
            action_label=_("Export"),
            title=_("Export"),
            label=_("Export"),
            description_fields=_("Select the export format"),
            serializer=ExportFileSerializer,
            instance_display=create_simple_display([["format"]]),
            weight=1,
        )



from .calendar_item import (
    CalendarItemModelSerializer,
    CalendarItemRepresentationSerializer,
)
from .conference_room import (
    BuildingModelSerializer,
    BuildingRepresentationSerializer,
    ConferenceRoomModelSerializer,
    ConferenceRoomRepresentationSerializer,
)



from rest_framework import serializers
from rest_framework.reverse import reverse

from .mixins import WBCoreSerializerFieldMixin
from .types import ReturnContentType, WBCoreType


class AbstractJSONField(serializers.JSONField):
    def __init__(self, serializer_class=None, flatten_field_names=None, *args, **kwargs):
        self.serializer_class = serializer_class
        if isinstance(flatten_field_names, list):
            self.flatten_field_names = flatten_field_names
        super().__init__(*args, **kwargs)


class JSONField(WBCoreSerializerFieldMixin, AbstractJSONField):
    field_type = WBCoreType.JSON.value


class JSONTableField(WBCoreSerializerFieldMixin, AbstractJSONField):
    field_type = WBCoreType.JSONTABLE.value


class JSONTextEditorField(WBCoreSerializerFieldMixin, serializers.JSONField):
    field_type = WBCoreType.TEXTEDITOR.value
    texteditor_content_type = ReturnContentType.JSON.value

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["content_type"] = self.texteditor_content_type
        return key, representation


class TemplatedJSONTextEditor(WBCoreSerializerFieldMixin, serializers.JSONField):
    field_type = WBCoreType.TEMPLATED_TEXTEDITOR.value

    def __init__(self, templates, default_editor_config=None, *args, **kwargs):
        self.templates = templates
        self.default_editor_config = default_editor_config
        super().__init__(*args, **kwargs)

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)

        if callable(self.templates):
            representation["template_url"] = self.templates(request)
        else:
            representation["template_url"] = reverse(self.templates, request=request)

        if self.default_editor_config:
            if callable(self.default_editor_config):
                representation["default_editor_config"] = self.default_editor_config(request)
            else:
                representation["default_editor_config"] = self.default_editor_config

        return key, representation



from datetime import date

from dynamic_preferences.preferences import Section
from dynamic_preferences.registries import global_preferences_registry
from dynamic_preferences.types import (
    DatePreference,
    IntegerPreference,
    StringPreference,
)

currency_section = Section("currency")


@global_preferences_registry.register
class DefaultCurrency(StringPreference):
    section = currency_section
    name = "default_currency"
    default = "USD"

    verbose_name = "Default Currency Used"


@global_preferences_registry.register
class DefaultStartDateHistoricalImport(DatePreference):
    section = currency_section
    name = "default_start_date_historical_import"
    default = date(2000, 1, 1)

    verbose_name = "Default Start Date"
    help_text = "Default start date in historical import"


@global_preferences_registry.register
class TimedeltaImportCurrencyFXRate(IntegerPreference):
    section = currency_section
    name = "timedelta_import_currency_fx_rates"
    default = 2

    verbose_name = "The Timedelta for the currency rate import windows"



# Generated by Django 4.1.9 on 2023-06-13 09:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("agenda", "0007_alter_calendaritem_options"),
    ]

    operations = [
        migrations.AlterField(
            model_name="calendaritem",
            name="item_type",
            field=models.CharField(blank=True, max_length=255, null=True, verbose_name="Type"),
        ),
    ]



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class CurrencyFXRatesCurrencyDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(fields=[dp.Field(key="date", label="Date"), dp.Field(key="value", label="Value")])



from django.db import models
from django.utils.translation import gettext_lazy as _

from wbcore.models import WBModel


class Display(WBModel):
    """A user modifiable display config that can be used to display process steps."""

    grid_template_areas = models.JSONField(verbose_name=_("Grid Fields"))
    name = models.CharField(max_length=128, verbose_name=_("Name"), unique=True)

    def __str__(self) -> str:
        return self.name

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:workflow:display"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:workflow:displayrepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"

    class Meta:
        verbose_name = _("Display")
        verbose_name_plural = _("Displays")



# Generated by Django 4.2.6 on 2024-01-25 13:10

import django.db.models.deletion
import django_fsm
import wbcore.contrib.icons.models
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("directory", "0006_employeremployeerelationship_position_name"),
        ("geography", "0001_initial"),
        ("agenda", "0008_alter_calendaritem_item_type"),
    ]

    operations = [
        migrations.CreateModel(
            name="League",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True, verbose_name="Name")),
                ("name", models.CharField(max_length=100, verbose_name="Name")),
                ("established_date", models.DateField(blank=True, null=True, verbose_name="Established Date")),
                ("website", models.URLField(blank=True, null=True, verbose_name="Website")),
                ("points_per_win", models.PositiveIntegerField(verbose_name="Points Per Win")),
                ("points_per_draw", models.PositiveIntegerField(default=1, verbose_name="Points Per Draw")),
                ("points_per_loss", models.PositiveIntegerField(default=0, verbose_name="Points Per Loss")),
            ],
            options={
                "verbose_name": "League",
                "verbose_name_plural": "Leagues",
            },
        ),
        migrations.CreateModel(
            name="Match",
            fields=[
                (
                    "calendaritem_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="agenda.calendaritem",
                    ),
                ),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True, verbose_name="Name")),
                ("date_time", models.DateTimeField(verbose_name="Date Time")),
                (
                    "status",
                    django_fsm.FSMField(
                        choices=[("SCHEDULED", "Scheduled"), ("ONGOING", "Ongoing"), ("FINISHED", "Finished")],
                        default="SCHEDULED",
                        max_length=50,
                        verbose_name="Status",
                    ),
                ),
                ("score_home", models.PositiveIntegerField(default=0, editable=False, verbose_name="Home Score")),
                ("score_away", models.PositiveIntegerField(default=0, editable=False, verbose_name="Away Score")),
                ("task_id", models.CharField(blank=True, editable=False, max_length=50)),
            ],
            options={
                "verbose_name": "Match",
                "verbose_name_plural": "Matches",
                "permissions": [("change_match_status", "Change Match Status")],
            },
            bases=("agenda.calendaritem", models.Model),
        ),
        migrations.CreateModel(
            name="Role",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("title", models.CharField(max_length=100, unique=True, verbose_name="Title")),
                ("slugify_title", models.CharField(blank=True, max_length=100, null=True, unique=True)),
            ],
            options={
                "verbose_name": "Role",
                "verbose_name_plural": "Roles",
            },
        ),
        migrations.CreateModel(
            name="Sport",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=100, unique=True, verbose_name="Name")),
                ("slugify_name", models.CharField(blank=True, max_length=100, null=True, unique=True)),
                ("rules", models.TextField(blank=True, max_length=100, null=True, verbose_name="Rules")),
                ("match_duration", models.PositiveIntegerField(verbose_name="Match Duration (min)")),
            ],
            options={
                "verbose_name": "Sport",
                "verbose_name_plural": "Sports",
            },
        ),
        migrations.CreateModel(
            name="SportPerson",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True, verbose_name="Name")),
                ("first_name", models.CharField(max_length=255, verbose_name="First Name")),
                ("last_name", models.CharField(max_length=255, verbose_name="Last Name")),
                (
                    "roles",
                    models.ManyToManyField(
                        blank=True, related_name="sport_persons", to="example_app.role", verbose_name="Roles"
                    ),
                ),
            ],
            options={
                "verbose_name": "Person",
                "verbose_name_plural": "Persons",
            },
        ),
        migrations.CreateModel(
            name="Stadium",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=100, unique=True, verbose_name="Name")),
                ("slugify_name", models.CharField(blank=True, max_length=100, null=True, unique=True)),
                ("capacity", models.PositiveIntegerField(blank=True, null=True, verbose_name="Capacity")),
                (
                    "city",
                    models.ForeignKey(
                        blank=True,
                        limit_choices_to={"level": 3},
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="stadiums",
                        to="geography.geography",
                        verbose_name="City",
                    ),
                ),
            ],
            options={
                "verbose_name": "Stadium",
                "verbose_name_plural": "Stadiums",
            },
        ),
        migrations.CreateModel(
            name="Team",
            fields=[
                (
                    "company_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="directory.company",
                    ),
                ),
                ("slugify_name", models.CharField(blank=True, max_length=100, null=True, unique=True)),
                ("founded_date", models.DateField(blank=True, null=True, verbose_name="Founded Date")),
                ("website", models.URLField(default="https://duckduckgo.com/", verbose_name="Website")),
                (
                    "city",
                    models.ForeignKey(
                        blank=True,
                        limit_choices_to={"level": 3},
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="teams",
                        to="geography.geography",
                        verbose_name="City",
                    ),
                ),
                (
                    "coach",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="coached_team",
                        to="example_app.sportperson",
                        verbose_name="Coach",
                    ),
                ),
                (
                    "home_stadium",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="teams_playing",
                        to="example_app.stadium",
                        verbose_name="Home Stadium",
                    ),
                ),
                (
                    "opponents",
                    models.ManyToManyField(
                        blank=True, through="example_app.Match", to="example_app.team", verbose_name="Opponents"
                    ),
                ),
            ],
            options={
                "verbose_name": "Team",
                "verbose_name_plural": "Teams",
            },
            bases=("directory.company",),
        ),
        migrations.CreateModel(
            name="Player",
            fields=[
                (
                    "sportperson_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="example_app.sportperson",
                    ),
                ),
                ("position", models.CharField(blank=True, max_length=50, null=True, verbose_name="Position")),
                (
                    "transfer_value",
                    models.DecimalField(
                        blank=True, decimal_places=2, max_digits=10, null=True, verbose_name="Market Value"
                    ),
                ),
            ],
            options={
                "verbose_name": "Player",
                "verbose_name_plural": "Players",
            },
            bases=("example_app.sportperson",),
        ),
        migrations.CreateModel(
            name="TeamResults",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("points", models.PositiveIntegerField(default=0, verbose_name="Points")),
                ("match_points_for", models.PositiveIntegerField(default=0, verbose_name="Match Points For")),
                ("match_points_against", models.PositiveIntegerField(default=0, verbose_name="Match Points Against")),
                ("match_point_difference", models.IntegerField(default=0, verbose_name="Match Point Difference")),
                ("wins", models.PositiveIntegerField(default=0, verbose_name="Wins")),
                ("draws", models.PositiveIntegerField(default=0, verbose_name="Draws")),
                ("losses", models.PositiveIntegerField(default=0, verbose_name="Losses")),
                ("form", models.CharField(default="-----", max_length=5, verbose_name="Form")),
                (
                    "league",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="results",
                        to="example_app.league",
                        verbose_name="League",
                    ),
                ),
                (
                    "team",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="results",
                        to="example_app.team",
                        verbose_name="Team",
                    ),
                ),
            ],
            options={
                "verbose_name": "Team Result",
                "verbose_name_plural": "Team Results",
            },
        ),
        migrations.AddField(
            model_name="match",
            name="away",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="away_matches",
                to="example_app.team",
                verbose_name="Away",
            ),
        ),
        migrations.AddField(
            model_name="match",
            name="home",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="home_matches",
                to="example_app.team",
                verbose_name="Home",
            ),
        ),
        migrations.AddField(
            model_name="match",
            name="league",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="matches",
                to="example_app.league",
                verbose_name="League",
            ),
        ),
        migrations.AddField(
            model_name="match",
            name="referee",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="refereed_matches",
                to="example_app.sportperson",
            ),
        ),
        migrations.AddField(
            model_name="match",
            name="sport",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="matches",
                to="example_app.sport",
                verbose_name="Sport",
            ),
        ),
        migrations.AddField(
            model_name="match",
            name="stadium",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="matches",
                to="example_app.stadium",
                verbose_name="Stadium",
            ),
        ),
        migrations.AddField(
            model_name="league",
            name="commissioner",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="commissioned_leagues",
                to="example_app.sportperson",
                verbose_name="Commissioner",
            ),
        ),
        migrations.AddField(
            model_name="league",
            name="country",
            field=models.ForeignKey(
                blank=True,
                limit_choices_to={"level": 1},
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="leagues",
                to="geography.geography",
                verbose_name="Country",
            ),
        ),
        migrations.AddField(
            model_name="league",
            name="sport",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="leagues",
                to="example_app.sport",
                verbose_name="Sport",
            ),
        ),
        migrations.AddField(
            model_name="league",
            name="teams",
            field=models.ManyToManyField(
                blank=True,
                related_name="leagues",
                through="example_app.TeamResults",
                to="example_app.team",
                verbose_name="Teams",
            ),
        ),
        migrations.CreateModel(
            name="EventType",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=100, verbose_name="Name")),
                (
                    "points",
                    models.PositiveIntegerField(
                        default=1,
                        help_text="Number of points awarded to a player's team per event",
                        verbose_name="Points",
                    ),
                ),
                (
                    "icon",
                    wbcore.contrib.icons.models.IconField(
                        default="sports_soccer", max_length=128, verbose_name="Icon"
                    ),
                ),
                (
                    "sport",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="event_types",
                        to="example_app.sport",
                        verbose_name="Sport",
                    ),
                ),
            ],
            options={
                "verbose_name": "Event Type",
                "verbose_name_plural": "Event Types",
            },
        ),
        migrations.CreateModel(
            name="Event",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("minute", models.PositiveIntegerField(verbose_name="Minute")),
                (
                    "event_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="events",
                        to="example_app.eventtype",
                        verbose_name="Event Type",
                    ),
                ),
                (
                    "match",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="events",
                        to="example_app.match",
                        verbose_name="Match",
                    ),
                ),
                (
                    "person",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="events",
                        to="example_app.sportperson",
                        verbose_name="Person",
                    ),
                ),
            ],
            options={
                "verbose_name": "Event",
                "verbose_name_plural": "Events",
            },
        ),
        migrations.AddConstraint(
            model_name="teamresults",
            constraint=models.UniqueConstraint(fields=("team", "league"), name="unique_team_league"),
        ),
        migrations.AddField(
            model_name="player",
            name="current_team",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="current_players",
                to="example_app.team",
                verbose_name="Current Team",
            ),
        ),
        migrations.AddField(
            model_name="player",
            name="former_teams",
            field=models.ManyToManyField(
                blank=True, related_name="former_players", to="example_app.team", verbose_name="Former Teams"
            ),
        ),
        migrations.AddConstraint(
            model_name="match",
            constraint=models.UniqueConstraint(fields=("home", "away", "date_time"), name="match_home_away_date_time"),
        ),
        migrations.AddConstraint(
            model_name="match",
            constraint=models.CheckConstraint(
                check=models.Q(("home", models.F("away")), _negated=True), name="check_match_home_away"
            ),
        ),
        migrations.AddConstraint(
            model_name="league",
            constraint=models.UniqueConstraint(fields=("name", "sport"), name="league_name_sport"),
        ),
        migrations.AddConstraint(
            model_name="eventtype",
            constraint=models.UniqueConstraint(fields=("name", "sport"), name="event_type_name_sport"),
        ),
        migrations.AddConstraint(
            model_name="event",
            constraint=models.UniqueConstraint(
                fields=("person", "match", "minute", "event_type"), name="event_person_match_minute_event_type"
            ),
        ),
    ]



from contextlib import suppress

from django.dispatch import receiver
from rest_framework.reverse import reverse
from wbcore.contrib.guardian.models.mixins import PermissionObjectModelMixin
from wbcore.contrib.icons.icons import WBIcon
from wbcore.metadata.configs.buttons import ButtonViewConfig
from wbcore.metadata.configs.buttons.buttons import WidgetButton
from wbcore.metadata.configs.buttons.enums import Button
from wbcore.signals.instance_buttons import add_extra_button


class PivotUserObjectPermissionButtonViewConfig(ButtonViewConfig):
    def get_create_buttons(self):
        return {
            Button.SAVE_AND_CLOSE.value,
        }


@receiver(add_extra_button)
def add_object_permission_button(sender, instance, request, view, pk=None, **kwargs):
    with suppress(AttributeError):
        if (
            (
                request.user.has_perm(f"administrate_{view.queryset.model._meta.model_name}")
                or request.user.is_superuser
            )
            and issubclass(view.queryset.model, PermissionObjectModelMixin)
            and pk is not None
        ):
            endpoint = reverse(
                "wbcore:guardian:pivoteduserobjectpermission-list",
                args=[view.get_content_type().id, pk],
                request=request,
            )
            return WidgetButton(endpoint=endpoint, label="Permissions", icon=WBIcon.LOCK.icon)



from django.db.models import Q
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from wbcore import viewsets
from wbcore.contrib.workflow.filters import TransitionFilter
from wbcore.contrib.workflow.models import Step, Transition, Workflow
from wbcore.contrib.workflow.serializers import TransitionModelSerializer
from wbcore.contrib.workflow.viewsets.display import TransitionDisplayConfig
from wbcore.contrib.workflow.viewsets.endpoints import TransitionEndpointConfig
from wbcore.contrib.workflow.viewsets.titles import TransitionTitleConfig
from wbcore.serializers import PrimaryKeyRelatedField


class TransitionModelViewSet(viewsets.ModelViewSet):
    LIST_DOCUMENTATION = "workflow/markdown/documentation/transition.md"
    queryset = Transition.objects.all()
    serializer_class = TransitionModelSerializer
    display_config_class = TransitionDisplayConfig
    title_config_class = TransitionTitleConfig
    endpoint_config_class = TransitionEndpointConfig
    search_fields = ("name",)
    ordering = ("name", "to_step__name")
    filterset_class = TransitionFilter
    ordering_fields = (
        "name",
        "from_step__name",
        "to_step__name",
    )

    def get_queryset(self):
        queryset = super().get_queryset().select_related("from_step", "to_step")
        if step_id := self.kwargs.get("step_id"):
            return queryset.filter(Q(from_step=step_id) | Q(to_step=step_id))
        if workflow_id := self.kwargs.get("workflow_id"):
            return queryset.filter(to_step__workflow=workflow_id)
        return queryset

    @cached_property
    def workflow(self) -> Workflow | None:
        if step_id := self.kwargs.get("step_id"):
            return Step.objects.get(id=step_id).workflow
        elif workflow_id := self.kwargs.get("workflow_id"):
            return Workflow.objects.get(id=workflow_id)

    def get_serializer_class(self):
        if self.workflow:

            class Serializer(TransitionModelSerializer):
                from_step = PrimaryKeyRelatedField(
                    queryset=Step.objects.filter(workflow=self.workflow), label=_("From"), required=False
                )
                to_step = PrimaryKeyRelatedField(
                    queryset=Step.objects.filter(workflow=self.workflow), label=_("To"), required=False
                )

            return Serializer
        return super().get_serializer_class()



import factory
from django.contrib.contenttypes.models import ContentType
from django.utils import timezone

from wbcore.contrib.directory.factories import PersonFactory
from wbcore.contrib.directory.models import Person
from wbcore.contrib.directory.serializers import PersonModelSerializer
from wbcore.contrib.workflow.models import Process, ProcessStep, Step


def _get_person_dict() -> dict:
    data = PersonModelSerializer(PersonFactory()).data
    # Serializer throws error when both fields are blank
    data["primary_email"] = None
    data["primary_telephone"] = None
    return data


class ProcessFactory(factory.django.DjangoModelFactory):
    id = factory.Faker("uuid4")
    workflow = factory.SubFactory("wbcore.contrib.workflow.factories.WorkflowFactory")
    started = factory.Faker("date_time", tzinfo=timezone.get_current_timezone())
    finished = factory.Faker("date_time", tzinfo=timezone.get_current_timezone())
    instance = factory.SubFactory("wbcore.contrib.directory.factories.PersonFactory")
    instance_id = factory.SelfAttribute("instance.pk")
    content_type = factory.LazyAttribute(lambda o: ContentType.objects.get_for_model(Person))
    preserved_instance = factory.Maybe(
        "preserve_instance_activated",
        yes_declaration=factory.LazyAttribute(lambda x: _get_person_dict()),
        no_declaration=None,
    )
    state = factory.Iterator(Process.ProcessState.values)

    class Meta:
        model = Process

    class Params:
        preserve_instance_activated = factory.SelfAttribute("workflow.preserve_instance")


class ProcessStepFactory(factory.django.DjangoModelFactory):
    id = factory.Faker("uuid4")
    process = factory.SubFactory(ProcessFactory)
    step = factory.SubFactory("wbcore.contrib.workflow.factories.RandomChildStepFactory")
    started = factory.Faker("date_time", tzinfo=timezone.get_current_timezone())
    finished = factory.Faker("date_time", tzinfo=timezone.get_current_timezone())
    assignee = factory.Maybe(
        "is_userstep",
        yes_declaration=factory.LazyAttribute(lambda x: x.step.get_casted_step().get_assigned_user()),
        no_declaration=None,
    )
    group = factory.Maybe(
        "is_userstep",
        yes_declaration=factory.LazyAttribute(lambda x: x.step.get_casted_step().get_assigned_group()),
        no_declaration=None,
    )
    state = factory.Iterator(ProcessStep.StepState.values)
    status = factory.SelfAttribute("step.status")
    error_message = factory.Maybe(
        "step_failed",
        yes_declaration=factory.Faker("text", max_nb_chars=128),
        no_declaration=None,
    )

    class Params:
        is_userstep = factory.LazyAttribute(lambda x: x.step.step_type == Step.StepType.USERSTEP)
        step_failed = factory.LazyAttribute(lambda x: x.state == ProcessStep.StepState.FAILED)

    class Meta:
        model = ProcessStep



import pytest
from django.contrib.auth.models import Permission


@pytest.mark.django_db
class TestCalendarItemTest:
    def test_to_ics(self, calendar_item_factory):
        calendar_item1 = calendar_item_factory()
        assert calendar_item1.to_ics(calendar_item1.period.upper, calendar_item1.period.lower) is None
        calendar_item2 = calendar_item_factory()
        assert calendar_item2.to_ics(calendar_item1.period.lower, calendar_item1.period.upper)

    def test_get_item_type(self, calendar_item_factory):
        calendar_item = calendar_item_factory()
        assert calendar_item.get_item_type() == "agenda.CalendarItem"

    @pytest.mark.parametrize("calendar_item__is_deletable", [True, False])
    def test_cannot_delete_calendar_item(self, calendar_item, user):
        assert not calendar_item.can_delete(
            user
        )  # as user is not a manger nor among the entities, he cannot delete the event

    @pytest.mark.parametrize("calendar_item__is_deletable", [True, False])
    def test_can_manager_delete_event(self, calendar_item, user):
        permission = Permission.objects.get(
            content_type__app_label="agenda", codename="administrate_confidential_items"
        )
        user.user_permissions.add(permission)
        assert calendar_item.can_delete(user) == calendar_item.is_deletable

    @pytest.mark.parametrize("calendar_item__is_deletable", [True, False])
    def test_can_participants_delete_event(self, calendar_item, user):
        calendar_item.entities.add(user.profile.id)
        assert calendar_item.can_delete(user) == calendar_item.is_deletable



from django.utils.translation import gettext as _

from wbcore.contrib.directory.models import Entry, SocialMediaContact
from wbcore.metadata.configs.titles import TitleViewConfig


class EmailContactTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Email Contact: {{address}}")

    def get_list_title(self):
        return _("Email Contacts")

    def get_create_title(self):
        return _("New Email Contact")


class AddressContactTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Address Contact")

    def get_list_title(self):
        return _("Address Contacts")

    def get_create_title(self):
        return _("New Address Contact")


class BankingContactTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Banking Contact: {{institute}}")

    def get_list_title(self):
        return _("Banking Contacts")

    def get_create_title(self):
        return _("New Banking Contact")


class EmailContactEntryTitleConfig(TitleViewConfig):
    def get_create_title(self):
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("New Email Contact for {entry}").format(entry=entry.computed_str)

    def get_list_title(self):
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("Email Contacts for {entry}").format(entry=entry.computed_str)


class AddressContactEntryTitleConfig(TitleViewConfig):
    def get_create_title(self):
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("New Address Contact for {entry}").format(entry=entry.computed_str)

    def get_list_title(self):
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("Address Contacts for {entry}").format(entry=entry.computed_str)


class TelephoneContactTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Telephone Contact: {{number}}")

    def get_list_title(self):
        return _("Telephone Contacts")

    def get_create_title(self):
        return _("New Telephone Contact")


class TelephoneContactEntryTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Telephone Contact: {{number}}")

    def get_list_title(self):
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("Telephone Contacts for {entry}").format(entry=entry.computed_str)

    def get_create_title(self):
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("New Telephone Contact for {entry}").format(entry=entry.computed_str)


class WebsiteContactEntryTitleConfig(TitleViewConfig):
    def get_instance_title(self):
        return _("Website Contact: {{url}}")

    def get_list_title(self):
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("Websites for {entry}").format(entry=entry.computed_str)

    def get_create_title(self):
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("New Website Contact for {entry}").format(entry=entry.computed_str)


class BankingContactEntryTitleConfig(TitleViewConfig):
    def get_list_title(self):
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("Banking Contacts for {entry}").format(entry=entry.computed_str)

    def get_create_title(self):
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("New Banking Contact for {entry}").format(entry=entry.computed_str)


class SocialMediaContactEntryTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str:
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("Social Media Accounts of {entry}").format(entry=entry.computed_str)

    def get_instance_title(self):
        try:
            contact_instance = self.view.get_object()
        except AssertionError:
            return _("Social Media Account")
        return _("{platform} Account of {entry}").format(
            platform=SocialMediaContact.Platforms[contact_instance.platform].value, entry=contact_instance.entry
        )

    def get_create_title(self) -> str:
        entry = Entry.all_objects.get(id=self.view.kwargs["entry_id"])
        return _("New Social Media Contact for {entry}").format(entry=entry.computed_str)



from functools import cached_property

from django.db.models.functions import Coalesce

from wbcore import viewsets
from wbcore.contrib.color.models import TransparencyMixin
from wbcore.contrib.directory.models import (
    ClientManagerRelationship,
    Company,
    Entry,
    Person,
)
from wbcore.contrib.i18n.viewsets import ModelTranslateMixin

from ...authentication.models import UserActivity
from ..filters import CompanyFilter, EntryFilter, PersonFilter, UserIsManagerFilter
from ..serializers import (
    BankModelSerializer,
    CompanyModelListSerializer,
    CompanyModelSerializer,
    CompanyRepresentationSerializer,
    EntryModelSerializer,
    EntryRepresentationSerializer,
    NewPersonModelSerializer,
    PersonModelListSerializer,
    PersonModelSerializer,
    PersonRepresentationSerializer,
)
from .buttons.entries import (
    CompanyModelButtonConfig,
    EntryModelButtonConfig,
    PersonModelButtonConfig,
)
from .display.entries import (
    BankModelDisplay,
    CompanyModelDisplay,
    EntryModelDisplay,
    PersonModelDisplay,
    UserIsManagerDisplayConfig,
)
from .endpoints.entries import (
    CompanyModelEndpointConfig,
    PersonModelEndpointConfig,
    UserIsManagerEndpointConfig,
)
from .mixins import EntryPermissionMixin
from .previews import EntryPreviewConfig
from .titles.entries import (
    CompanyModelTitleConfig,
    PersonModelTitleConfig,
    UserIsManagerTitleConfig,
)


class EntryRepresentationViewSet(EntryPermissionMixin, viewsets.RepresentationViewSet):
    min_rank = 0.25
    queryset = Entry.objects.annotate_all()
    serializer_class = EntryRepresentationSerializer
    filterset_class = EntryFilter
    ordering = ("entry_type", "id")
    ordering_fields = ("computed_str",)
    search_fields = ("computed_str", "slugify_computed_str", "emails__address")


class PersonRepresentationViewSet(EntryRepresentationViewSet):
    queryset = Person.objects.annotate_all()
    serializer_class = PersonRepresentationSerializer
    filterset_class = PersonFilter

    ordering_fields = ("last_name", "first_name")
    ordering = ("last_name", "first_name", "id")


class PersonInChargeRepresentationViewSet(PersonRepresentationViewSet):
    def get_queryset(self):
        return super().get_queryset().filter(clients__isnull=False).distinct()


class CompanyRepresentationViewSet(EntryRepresentationViewSet):
    queryset = Company.objects.annotate_all()
    serializer_class = CompanyRepresentationSerializer
    filterset_class = CompanyFilter
    ordering_fields = ("name",)
    ordering = ("name", "id")


class EntryModelViewSet(EntryPermissionMixin, TransparencyMixin, viewsets.ModelViewSet):
    min_rank = 0.25
    search_fields = ("computed_str", "slugify_computed_str", "emails__address")
    ordering = ["computed_str", "id"]
    ordering_fields = (
        "computed_str",
        "primary_telephone",
        "primary_email",
        "primary_address",
    )
    preview_config_class = EntryPreviewConfig
    queryset = Entry.objects.annotate_all()
    serializer_class = EntryModelSerializer
    filterset_class = EntryFilter

    button_config_class = EntryModelButtonConfig
    display_config_class = EntryModelDisplay

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .prefetch_related(
                "addresses",
                "relationship_managers",
                "social_media",
            )
        )

    @cached_property
    def is_internal_user(self):
        return self.request.user.profile and self.request.user.profile.is_internal


class PersonModelViewSet(ModelTranslateMixin, EntryModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/person.md"
    button_config_class = PersonModelButtonConfig
    display_config_class = PersonModelDisplay
    endpoint_config_class = PersonModelEndpointConfig
    filterset_class = PersonFilter
    ordering = ["-activity_heat", "id"]
    ordering_fields = (
        "name",
        "primary_employer_repr",
        "customer_status",
        "position_in_company",
        "primary_telephone",
        "tier",
        "primary_manager_repr",
        "last_event",
        "last_event_period_endswith",
        "activity_heat",
    )

    queryset = Person.objects.annotate_all()
    serializer_class = PersonModelSerializer
    title_config_class = PersonModelTitleConfig

    def get_serializer_class(self):
        if self.get_action() in ["list", "list-metadata"]:
            return PersonModelListSerializer
        elif "pk" not in self.kwargs:
            return NewPersonModelSerializer
        return super().get_serializer_class()

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(last_connection=Coalesce(UserActivity.get_latest_login_datetime_subquery(), None))
        )


class CompanyModelViewSet(EntryModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/company.md"
    button_config_class = CompanyModelButtonConfig
    display_config_class = CompanyModelDisplay
    endpoint_config_class = CompanyModelEndpointConfig
    filterset_class = CompanyFilter
    ordering_fields = (
        "name",
        "type",
        "tier",
        "customer_status",
        "relationship_managers",
        "last_event",
        "last_event_period_endswith",
        "activity_heat",
    )
    ordering = ["name", "id"]
    queryset = Company.objects.annotate_all().select_related("entry_ptr", "customer_status", "type")
    serializer_class = CompanyModelSerializer
    title_config_class = CompanyModelTitleConfig

    def get_serializer_class(self):
        if self.get_action() in ["list", "list-metadata"]:
            return CompanyModelListSerializer
        return super().get_serializer_class()


class BankModelViewSet(EntryModelViewSet):
    queryset = Company.objects.annotate_all()
    display_config_class = BankModelDisplay
    serializer_class = BankModelSerializer


class UserIsManagerViewSet(EntryModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/userismanager.md"
    display_config_class = UserIsManagerDisplayConfig
    title_config_class = UserIsManagerTitleConfig
    endpoint_config_class = UserIsManagerEndpointConfig
    search_fields = ("computed_str", "slugify_computed_str")
    filterset_class = UserIsManagerFilter

    def get_queryset(self):
        client_ids = ClientManagerRelationship.objects.filter(
            relationship_manager=self.request.user.profile.id,
            status__in=[ClientManagerRelationship.Status.APPROVED, ClientManagerRelationship.Status.PENDINGREMOVE],
        ).values_list("client_id", flat=True)
        return super().get_queryset().filter(id__in=client_ids)



from unittest.mock import patch

import pytest
from faker import Faker
from rest_framework.viewsets import ViewSet

from wbcore.filters.fields import CharFilter

fake = Faker()


class TestWBCoreFilterMixin:
    @pytest.fixture()
    def view(self):
        return ViewSet()

    @pytest.mark.parametrize("field_name", ["Foo"])
    def test_key_column_name(self, field_name):
        filter_field = CharFilter(column_field_name=field_name, field_name=field_name)
        assert filter_field.key == field_name

    @pytest.mark.parametrize("field_name", ["Foo"])
    def test_key_field_name(self, field_name):
        filter_field = CharFilter(field_name=field_name)
        assert filter_field.key == field_name

    @pytest.mark.parametrize("label", ["Foo"])
    def test_get_label_with_label(self, label):
        filter_field = CharFilter(label=label)
        assert filter_field.get_label() == label

    @pytest.mark.parametrize("field_name1, field_name2", [("Foo", "Bar")])
    def test_get_label_without_label(self, field_name1, field_name2):
        field_name = field_name1 + "_" + field_name2
        filter_field = CharFilter(field_name=field_name)
        assert filter_field.get_label() == field_name1.title() + " " + field_name2.title()

    @pytest.mark.parametrize("initial", ["Foo"])
    def test__get_initial_with_callable(self, rf, initial, view):
        filter_field = CharFilter(initial=lambda f, r, v: initial)
        assert filter_field._get_initial(rf, view) == initial

    @pytest.mark.parametrize("initial", ["Foo"])
    def test__get_initial_with_callable_str(self, rf, initial, view):
        class CustomFilterField(CharFilter):
            def custom_initial(self, *args):
                return initial

        filter_field = CustomFilterField(initial="custom_initial")
        assert filter_field._get_initial(rf, view) == initial

    @pytest.mark.parametrize("initial", ["Foo"])
    def test__get_initial_with_initial(self, rf, initial, view):
        filter_field = CharFilter(initial=initial)
        assert filter_field._get_initial(rf, view) == initial

    @pytest.mark.parametrize("name", ["Foo"])
    def test_get_representation(self, name, rf, view):
        filter_field = CharFilter(field_name=name)
        request = rf.get("")
        rep, _ = filter_field.get_representation(request, name, view)
        for key in ["label", "key", "label_format"]:
            assert key in rep

    @pytest.mark.parametrize("name, initial", [("Foo", "Bar")])
    def test_get_representation_request_initial(self, name, initial, rf, view):
        filter_field = CharFilter(field_name=name)
        rf.GET = {name: initial}
        _, le = filter_field.get_representation(rf, name, view)
        assert le["input_properties"]["initial"] == initial

    @pytest.mark.parametrize("name, initial", [("Foo", "Bar")])
    def test_get_representation_field_initial(self, name, initial, rf, view):
        filter_field = CharFilter(field_name=name, initial=initial)
        request = rf.get("")
        _, le = filter_field.get_representation(request, name, view)
        assert le["input_properties"]["initial"] == initial

    @pytest.mark.parametrize("name, initial, help_text", [("Foo", "Bar", "Sesquipedalophobie")])
    def test_get_help_text(self, name, initial, help_text, rf, view):
        request = rf.get("")

        # assert that no field nor help text nor label returns an empty help_text
        filter_field = CharFilter(field_name=name, initial=initial, help_text=None)
        assert filter_field.get_representation(request, name, view)[0]["help_text"] is None

        # assert that no field nor help text but a label returns an "Filter by {{label}}"
        filter_field = CharFilter(field_name=name, initial=initial, help_text=None, label=help_text)
        assert filter_field.get_representation(request, name, view)[0]["help_text"] == f"Filter by {help_text}"

        filter_field = CharFilter(field_name=name, initial=initial, help_text=help_text)
        assert filter_field.get_representation(request, name, view)[0]["help_text"] == help_text

    @patch("wbcore.filters.mixins.get_model_field")
    @pytest.mark.parametrize("name, initial, text", [("Foo", "Bar", "Sesquipedalophobie")])
    def test_get_help_text_from_model_help_text(self, mock_fct, name, initial, text, rf, view):
        class Field:
            help_text = text

        class dummy_parent:
            class _meta:
                model = "tmp"

        mock_fct.return_value = Field

        filter_field = CharFilter(field_name=name, initial=initial, help_text=None)
        filter_field.parent = dummy_parent()
        request = rf.get("")
        assert filter_field.get_representation(request, name, view)[0]["help_text"] == text



from dataclasses import dataclass, field

from wbcore.metadata.configs.display.instance_display.styles import Style
from wbcore.metadata.configs.display.instance_display.utils import grid_definition

from .inlines import Inline, SerializedInline
from .sections import Section, SerializedSection

type SerializedLayout = dict[str, str | list[SerializedInline] | list[SerializedSection]]


@dataclass
class Layout:
    """A layout contains all the information on how a form is build.

    A form is rendered based on a grid, which is closely modelled after a CSS Grid. Essentially, we can define
    rows and columns and places fields on this grid, either in one cell, or spanning multiple rows and columns.

    In the future we will add all other features of css grid to ensure full control over the grid.

    Attributes:
        grid_template_areas: A 2D string Array containing the areas on the grid. The strings are the keys references to all fields
        grid_template_columns: A string array containing the column definitions for the width of each column
        grid_template_rows: A string array containing the row definitions for the height of each row
        sections: A list of sections that can be positioned in `grid_template_areas`
        inlines: A list of inlines that can be positioned in `grid_template_areas`
        gap: A string specifying the gap between all rows and columns. Should be used with a helper method like `Style.px(10)`

    """

    grid_template_areas: list[list[str]]
    grid_auto_columns: str | None = None
    grid_template_columns: list[str] | None = None
    grid_auto_rows: str = Style.MIN_CONTENT
    grid_template_rows: list[str] | None = None

    sections: list[Section] = field(default_factory=list)
    inlines: list[Inline] = field(default_factory=list)

    gap: str = Style.px(10)
    row_gap: str | None = None
    column_gap: str | None = None

    def serialize(self, key_prefix: str | None = None) -> SerializedLayout:
        """Serializes this `Layout`
        Attributes:
            key_prefix(Optional): if specified, append a prefix to underlying inlines key
        Returns:
            A dictionairy containing all serialized fields

        """
        serialized_layout = {
            "gridTemplateAreas": " ".join([f"'{' '.join(row)}'" for row in self.grid_template_areas]),
            "sections": [section.serialize(key_prefix=key_prefix) for section in self.sections],
            "inlines": [inline.serialize(key_prefix=key_prefix) for inline in self.inlines],
            "gap": self.gap,
        }

        if self.row_gap:
            serialized_layout["rowGap"] = self.row_gap

        if self.column_gap:
            serialized_layout["columnGap"] = self.column_gap

        if self.grid_auto_columns:
            serialized_layout["gridAutoColumns"] = self.grid_auto_columns

        if self.grid_template_columns:
            serialized_layout["gridTemplateColumns"] = grid_definition(*self.grid_template_columns)

        if self.grid_auto_rows:
            serialized_layout["gridAutoRows"] = self.grid_auto_rows

        if self.grid_template_rows:
            serialized_layout["gridTemplateRows"] = grid_definition(*self.grid_template_rows)

        return serialized_layout



from typing import Optional

from django.utils.translation import gettext_lazy as _

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class CurrencyDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label="Title"),
                dp.Field(key="symbol", label="Sybmol"),
                dp.Field(key="key", label="Key"),
                dp.Field(key="rates_sparkline", label="Rates"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["title", "symbol", "key"], [repeat_field(3, "currencyfxrates_section")]],
            [
                create_simple_section(
                    "currencyfxrates_section",
                    _("Rates"),
                    [["currency_currencyfxrates"]],
                    "currency_currencyfxrates",
                    collapsed=False,
                )
            ],
        )



from django.shortcuts import get_object_or_404
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT
from rest_framework.views import APIView

from .models import AppliedPreset


class PresetAPIView(APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request: Request, display_identifier_path: str) -> Response:
        preset = get_object_or_404(
            AppliedPreset,
            display_identifier_path=display_identifier_path,
            user=request.user,
        )

        return Response(preset.display, status=HTTP_200_OK)

    def delete(self, request: Request, display_identifier_path: str) -> Response:
        preset = get_object_or_404(
            AppliedPreset,
            display_identifier_path=display_identifier_path,
            user=request.user,
        )

        preset.delete()

        return Response(status=HTTP_204_NO_CONTENT)

    def post(self, request: Request, display_identifier_path: str) -> Response:
        defaults = {}

        if display := self.request.data.get("display", None):
            defaults["display"] = display

        if preset := self.request.data.get("preset", None):
            defaults["preset"] = preset

        AppliedPreset.objects.update_or_create(
            user=request.user,
            display_identifier_path=display_identifier_path,
            defaults=defaults,
        )
        return Response(status=HTTP_201_CREATED)



import json
from functools import reduce


def enumerated_string_join(array):
    """Generates a human readable string out of a list"""

    if len(array) > 0:
        casted_array = [str(i) for i in array]
        return (
            " and ".join([", ".join(casted_array[:-1]), casted_array[-1]])
            if len(casted_array) > 1
            else casted_array[0]
        )
    return ""


def format_number(number, decimal: int = 2, **kwargs) -> float | None:
    """
    utility function used to serialize an aggregate to a json compatible value
    Args:
        number: Number to serialize
        **kwargs: backward compatibility keyword argument

    Returns:

    """
    try:
        return float(round(number, decimal))
    except TypeError:
        return None


class ReferenceIDMixin:
    @property
    def reference_id(self):
        return f"{self.id:06}"


def convert_str_to_boolean(fake_bool: str | bool) -> bool:
    """
    Converts any bool or str to a boolean - can throw JSONDecodeError
    """
    return json.loads(str(fake_bool).lower())


def snake_case_to_human(s: str) -> str:
    words = s.split("_")
    human_readable = " ".join(word.capitalize() for word in words)
    return human_readable


def camel_to_snake_case(s: str) -> str:
    """
    Convert a string as CamelCase to snake_case
    """
    return reduce(lambda x, y: x + ("_" if y.isupper() else "") + y, s).lower()


def get_aggregate_symbol(aggregate: str) -> str:
    match aggregate.lower():
        case "avg":
            return "μ"
        case "count":
            return "#"
        case "max":
            return "⋁"
        case "min":
            return "⋀"
        case "stddev":
            return "σ"
        case "sum":
            return "Σ"
        case "variance":
            return "σ2"
        case _:
            return aggregate



from functools import lru_cache

from django.db.models import QuerySet


@lru_cache
def mapping(queryset: QuerySet, mapping_field: str = "key", id_field: str = "id", **filter_kwargs) -> dict:
    return dict(queryset.filter(**filter_kwargs).values_list(mapping_field, id_field))



from wbcore.test.mixins import TestViewSet
from wbcore.test.utils import get_or_create_superuser


class TestPandasView(TestViewSet):
    def test_get_request(self):
        if self.factory is None:
            self._raise_warning_factory("test_get_request")
        else:
            obj, request, kwargs, _ = self._get_mixins_data()
            request.user = get_or_create_superuser()
            request.query_params = dict()
            request.parser_context = {}
            self.mvs.kwargs = kwargs
            response = self.mvs.as_view({"get": "list"})(request, **kwargs)
            assert response.status_code == 200
            assert "results" in response.data
            assert "aggregates" in response.data
            assert "messages" in response.data

    def execute_test(self):
        self.test_get_request()



from wbcore.menus import ItemPermission, MenuItem

TAG_MENUITEM = MenuItem(
    label="Tag",
    endpoint="wbcore:tags:tag-list",
    add=MenuItem(label="New Tag", endpoint="wbcore:tags:tag-list"),
    permission=ItemPermission(permissions=["tags.change_tag"]),
)

TAGGROUP_MENUITEM = MenuItem(
    label="Tag Group",
    endpoint="wbcore:tags:taggroup-list",
    add=MenuItem(label="New Tag Group", endpoint="wbcore:tags:taggroup-list"),
    permission=ItemPermission(permissions=["tags.change_taggroup"]),
)



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig
from wbcore.metadata.configs.preview import PreviewViewConfig


class TagGroupDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(fields=[dp.Field(key="title", label="Title")])

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [["title"], ["tags_sections"]],
            [create_simple_section("tags_sections", "Tags", [["tags"]], "tags", collapsed=True)],
        )


class TagDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="title", label="Title"),
                dp.Field(key="color", label="Color"),
                dp.Field(key="slug", label="Slug"),
                dp.Field(key="groups", label="Groups"),
                dp.Field(key="content_type", label="Content Type"),
                dp.Field(key="description", label="Description"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(2, "title"), "slug"],
                ["groups", "color", "content_type"],
                [repeat_field(3, "description")],
            ]
        )


class TagPreviewConfig(PreviewViewConfig):
    def get_display(self):
        return create_simple_display([["title", "groups", "description"]])



from celery import shared_task
from wbcore.contrib.guardian.models.mixins import PermissionObjectModelMixin
from wbcore.utils.itertools import get_inheriting_subclasses


@shared_task
def reload_permissions_as_task(prune_existing: bool | None = True, force_pruning: bool | None = False):
    for subclass in get_inheriting_subclasses(PermissionObjectModelMixin):
        for instance in subclass.objects.iterator():
            instance.reload_permissions(prune_existing=prune_existing, force_pruning=force_pruning)



from rest_framework import serializers
from rest_framework.generics import ListAPIView
from rest_framework.test import APIRequestFactory

from wbcore.pagination import LimitOffsetPagination

factory = APIRequestFactory()


class TestEndlessPaginationMixin:
    def setup_method(self):
        class PassThroughSerializer(serializers.Serializer):
            def to_representation(self, instance):
                return instance

        self.view = ListAPIView.as_view(
            serializer_class=PassThroughSerializer,
            queryset=range(1, 101),
            pagination_class=LimitOffsetPagination,
            authentication_classes=[],
            permission_classes=[],
        )

    def test_paginate_queryset_with_pagination(self):
        request = factory.get("/")
        response = self.view(request)
        assert len(response.data["results"]) == 25

    def test_paginate_queryset_without_pagination(self):
        request = factory.get("/", {"disable_pagination": "true"})
        response = self.view(request)
        assert len(response.data) == 100



from django.utils.translation import gettext as _

from wbcore.metadata.configs.titles import TitleViewConfig


class ProcessTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Processes")

    def get_instance_title(self):
        return _("Process")


class ProcessStepTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Process Steps")

    def get_instance_title(self):
        return _("Process Step")


class AssignedProcessStepTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return _("Assigned Process Steps")

    def get_instance_title(self):
        return _("Assigned Process Step")



from django.db.models.signals import m2m_changed

from wbcore.dispatch import receiver_inherited_through_models

from .models.mixins import TagModelMixin


@receiver_inherited_through_models(m2m_changed, sender=TagModelMixin, through_field_name="tags")
def pre_add_tags(sender, instance, action, pk_set, **kwargs):
    from django.contrib.contenttypes.models import ContentType

    from wbcore.contrib.tags.models.tags import Tag

    if action == "post_add" and pk_set:
        content_type = ContentType.objects.get_for_model(instance)
        for tag_id in pk_set:
            tag = Tag.objects.get(id=tag_id)
            if tag.content_type and tag.content_type != content_type:
                instance.tags.remove(tag)



import pytest
from django.apps import apps
from django.db.backends.base.base import BaseDatabaseWrapper
from pytest_mock import MockerFixture
from rest_framework.test import APIRequestFactory, APIClient
from django.db.models.signals import pre_migrate
from pytest_factoryboy import register
from wbcore.contrib.geography.tests.signals import app_pre_migration
from wbcore.contrib.workflow.factories import (
    ConditionFactory,
    DataFactory,
    DataValueFactory,
    DecisionStepFactory,
    EmailStepFactory,
    FinishStepFactory,
    JoinStepFactory,
    ProcessFactory,
    ProcessStepFactory,
    RandomChildStepFactory,
    ScriptStepFactory,
    SplitStepFactory,
    StartStepFactory,
    StepFactory,
    TransitionFactory,
    UserStepFactory,
    WorkflowFactory,
)
from wbcore.contrib.authentication.factories import PermissionFactory

from wbcore.contrib.workflow.models import (
    ProcessStep,
    Step,
)

from wbcore.contrib.workflow.serializers import (
    AssignedProcessStepSerializer,
    ConditionModelSerializer,
    DataModelSerializer,
    DecisionStepModelSerializer,
    ProcessModelSerializer,
    ProcessStepModelSerializer,
    TransitionModelSerializer,
    WorkflowModelSerializer,
)

from wbcore.tests.conftest import *

register(WorkflowFactory)
register(EmailStepFactory)
register(TransitionFactory)
register(DataFactory)
register(ProcessStepFactory)
register(UserStepFactory)
register(ConditionFactory)
register(RandomChildStepFactory)
register(FinishStepFactory)
register(ProcessFactory)
register(DecisionStepFactory)
register(SplitStepFactory)
register(JoinStepFactory)
register(ScriptStepFactory)
register(StepFactory)
register(DataValueFactory)
register(StartStepFactory)


# ============================== General Fixtures ==============================


@pytest.fixture
def api_client(super_user):
    client = APIClient()
    client.force_authenticate(user=super_user)
    return client


@pytest.fixture
def request_factory():
    return APIRequestFactory()


@pytest.fixture
def super_user(django_db_setup, django_db_blocker):
    return UserFactory(is_superuser=True)


@pytest.fixture
def user(django_db_setup, django_db_blocker):
    with django_db_blocker.unblock():
        return UserFactory(is_superuser=False)


@pytest.fixture()
def user_step():
    return UserStepFactory(step_type=Step.StepType.USERSTEP)


@pytest.fixture
def permission(django_db_setup, django_db_blocker):
    with django_db_blocker.unblock():
        return PermissionFactory()


# ============================== Condition Fixtures ==============================


@pytest.fixture
def condition(conditions):
    return conditions[0]


@pytest.fixture
def condition_build():
    transition = TransitionFactory()
    instance = ConditionFactory.build(transition=transition)
    return ConditionModelSerializer(instance).data


@pytest.fixture
def conditions():
    return ConditionFactory.create_batch(3)


# ============================== Data Fixtures ==============================


@pytest.fixture
def data():
    return DataFactory.create_batch(3)


@pytest.fixture
def data_build(workflow):
    instance = DataFactory.build(workflow=workflow)
    return DataModelSerializer(instance).data


@pytest.fixture
def singular_data(data):
    return data[0]


# ============================== Decision Step Fixtures ==============================


@pytest.fixture
def decision_step(decision_steps):
    return decision_steps[0]


@pytest.fixture
def decision_step_build(workflow):
    instance = DecisionStepFactory.build(workflow=workflow)
    return DecisionStepModelSerializer(instance).data


@pytest.fixture
def decision_steps():
    return DecisionStepFactory.create_batch(3)


# ============================== Process Fixtures ==============================


@pytest.fixture
def process(processes):
    return processes[0]


@pytest.fixture
def process_build():
    instance = ProcessFactory.build()
    return ProcessModelSerializer(instance).data


@pytest.fixture
def processes():
    return ProcessFactory.create_batch(3)


# ============================== Process Step Fixtures ==============================


@pytest.fixture
def process_step(process_steps):
    return process_steps[0]


@pytest.fixture
def process_steps():
    return ProcessStepFactory.create_batch(3)


@pytest.fixture
def process_steps_build():
    instance = ProcessStepFactory.build()
    return ProcessStepModelSerializer(instance).data


@pytest.fixture
def mocked_process_step(mocker: MockerFixture):
    return mocker.MagicMock(spec=ProcessStep)


# ============================== Assigned Process Step Fixtures ==============================


@pytest.fixture
def assigned_process_step(assigned_process_steps):
    return assigned_process_steps[0]


@pytest.fixture
def assigned_process_steps(super_user, user_step):
    return ProcessStepFactory.create_batch(
        3,
        step=user_step,
        state=ProcessStep.StepState.ACTIVE,
        assignee=super_user,
    )


@pytest.fixture
def assigned_process_steps_build(super_user, user_step):
    instance = ProcessStepFactory.build(
        step=user_step, state=ProcessStep.StepState.ACTIVE, assignee=super_user
    )
    return AssignedProcessStepSerializer(instance).data


# ============================== Transition Fixtures ==============================


@pytest.fixture
def transition(transitions):
    return transitions[0]


@pytest.fixture
def transition_build():
    to_step = FinishStepFactory()
    from_step = StartStepFactory(workflow=to_step.workflow)
    instance = TransitionFactory.build(to_step=to_step, from_step=from_step)
    return TransitionModelSerializer(instance).data


@pytest.fixture
def transitions():
    return TransitionFactory.create_batch(3)


# ============================== Workflow Fixtures ==============================


@pytest.fixture
def workflow(workflows):
    return workflows[0]


@pytest.fixture
def workflow_build():
    instance = WorkflowFactory.build()
    return WorkflowModelSerializer(instance).data


@pytest.fixture
def workflows():
    return WorkflowFactory.create_batch(3)


@pytest.fixture(autouse=True, scope="session")
def django_test_environment(django_test_environment):
    from django.apps import apps

    get_models = apps.get_models

    for m in [m for m in get_models() if not m._meta.managed]:
        m._meta.managed = True


pre_migrate.connect(app_pre_migration, sender=apps.get_app_config("geography"))



from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse

from wbcore import serializers as wb_serializers

from ..models import Entry


class EntryRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.SerializerMethodField()
    _detail_preview = wb_serializers.HyperlinkField(reverse_name="wbcore:directory:entry-detail")
    primary_email = wb_serializers.CharField(default="", label=_("Primary Email"), allow_null=True)
    primary_telephone = wb_serializers.TelephoneField(default="", label=_("Primary Telephone"), allow_null=True)

    def get__detail(self, obj):
        if obj.is_company:
            return reverse("wbcore:directory:company-detail", args=[obj.id], request=self.context.get("request"))
        else:
            return reverse("wbcore:directory:person-detail", args=[obj.id], request=self.context.get("request"))

    class Meta:
        model = Entry
        fields = (
            "id",
            "computed_str",
            "_detail",
            "_detail_preview",
            "primary_email",
            "primary_telephone",
        )


class EntryUnlinkedRepresentationSerializer(EntryRepresentationSerializer):
    class Meta(EntryRepresentationSerializer.Meta):
        model = Entry
        fields = ("id", "computed_str")



from contextlib import suppress

from django.conf import settings
from django.contrib.auth.models import Group
from django.db.models import QuerySet
from django.utils.functional import cached_property

from wbcore.contrib.authentication.models import User
from wbcore.utils.importlib import import_from_dotted_path


class UserBackendRegistry:
    def __init__(self):
        internal_users_backend_path = getattr(settings, "USER_BACKEND", "wbcore.permissions.backend.UserBackend")
        internal_users_backend_class = import_from_dotted_path(internal_users_backend_path)
        self.backend = internal_users_backend_class()

    @cached_property
    def internal_groups(self) -> QuerySet[Group]:
        return self.backend.get_internal_groups().all()

    @cached_property
    def internal_users(self) -> QuerySet[User]:
        return self.backend.get_internal_users().all()

    def reset_cache(self):
        with suppress(AttributeError):
            del self.internal_users
        with suppress(AttributeError):
            del self.internal_groups


user_registry = UserBackendRegistry()



from django.db.models import CharField, F, OuterRef, Q, Subquery, Value
from django.db.models.functions import Concat
from django.utils.functional import cached_property
from dynamic_preferences.registries import global_preferences_registry
from rest_framework import filters
from rest_framework.permissions import IsAuthenticated
from rest_fuzzysearch.search import RankedFuzzySearchFilter
from reversion.views import RevisionMixin

from wbcore.contrib.directory.models import (
    ClientManagerRelationship,
    EmailContact,
    EmployerEmployeeRelationship,
    Relationship,
    RelationshipType,
    TelephoneContact,
)
from wbcore.filters import DjangoFilterBackend
from wbcore.viewsets import ModelViewSet, ReadOnlyModelViewSet, RepresentationViewSet

from ..filters import ClientManagerFilter, RelationshipEntryFilter, RelationshipFilter
from ..serializers import (
    ClientManagerModelSerializer,
    ClientManagerRelationshipRepresentationSerializer,
    EmployerEmployeeRelationshipSerializer,
    RelationshipModelSerializer,
    RelationshipRepresentationSerializer,
    RelationshipTypeModelSerializer,
    RelationshipTypeRepresentationSerializer,
    UserIsClientModelSerializer,
)
from .buttons import EmployerEmployeeRelationshipButtonConfig
from .display import (
    ClientManagerModelDisplay,
    EmployeeEmployerDisplayConfig,
    EmployerEmployeeDisplayConfig,
    EmployerEmployeeRelationshipDisplayConfig,
    RelationshipDisplayConfig,
    RelationshipEntryDisplay,
    RelationshipTypeDisplayConfig,
    UserIsClientDisplayConfig,
)
from .endpoints import (
    ClientManagerEndpoint,
    EmployeeEmployerEndpointConfig,
    EmployerEmployeeEndpointConfig,
    RelationshipEntryModelEndpoint,
    UserIsClientEndpointConfig,
)
from .titles import (
    ClientManagerTitleConfig,
    EmployerEmployeeRelationshipTitleConfig,
    RelationshipEntryModelTitleConfig,
    RelationshipTypeModelTitleConfig,
    UserIsClientTitleConfig,
)


class RelationshipRepresentationViewSet(RepresentationViewSet):
    queryset = Relationship.objects.all()
    serializer_class = RelationshipRepresentationSerializer


class RelationshipModelViewSet(RevisionMixin, ModelViewSet):
    queryset = Relationship.objects.select_related(
        "relationship_type",
        "from_entry",
        "to_entry",
    )
    serializer_class = RelationshipModelSerializer

    display_config_class = RelationshipDisplayConfig

    search_fields = ("computed_str",)
    filterset_class = RelationshipFilter
    ordering = ["relationship_type", "to_entry", "id"]


class RelationshipModelEntryViewSet(RelationshipModelViewSet):
    title_config_class = RelationshipEntryModelTitleConfig
    display_config_class = RelationshipEntryDisplay
    endpoint_config_class = RelationshipEntryModelEndpoint
    filterset_class = RelationshipEntryFilter

    def get_queryset(self):
        _id = self.kwargs["entry_id"]
        return super().get_queryset().filter(Q(from_entry=_id)).distinct()


class RelationshipTypeRepresentationViewSet(RepresentationViewSet):
    queryset = RelationshipType.objects.all()
    serializer_class = RelationshipTypeRepresentationSerializer

    search_fields = ("title",)


class RelationshipTypeModelViewSet(ModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/relationshiptype.md"
    title_config_class = RelationshipTypeModelTitleConfig

    display_config_class = RelationshipTypeDisplayConfig
    serializer_class = RelationshipTypeModelSerializer
    queryset = RelationshipType.objects.select_related("counter_relationship")
    ordering_fields = (
        "title",
        "counter_relationship",
    )
    ordering = ("title",)


class EmployerEmployeeRelationshipModelViewSet(ModelViewSet):
    serializer_class = EmployerEmployeeRelationshipSerializer

    display_config_class = EmployerEmployeeRelationshipDisplayConfig
    button_config_class = EmployerEmployeeRelationshipButtonConfig
    title_config_class = EmployerEmployeeRelationshipTitleConfig
    ordering_fields = ("primary", "position__title", "position_name")

    filterset_fields = {"position": ["exact"], "primary": ["exact"], "position_name": ["exact", "icontains"]}
    queryset = EmployerEmployeeRelationship.objects.all().prefetch_related("employee", "employer", "position")


class EmployeeEmployerModelViewSet(EmployerEmployeeRelationshipModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/employeecompany.md"

    endpoint_config_class = EmployeeEmployerEndpointConfig
    display_config_class = EmployeeEmployerDisplayConfig

    ordering = ["employee__computed_str", "id"]
    ordering_fields = ["employee__computed_str"]
    search_fields = ["employee__computed_str"]
    filterset_fields = {"employee": ["exact"]}

    def get_queryset(self):
        return super().get_queryset().filter(employer=self.kwargs["employer_id"])


class EmployerEmployeeModelViewSet(EmployerEmployeeRelationshipModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/employerperson.md"

    endpoint_config_class = EmployerEmployeeEndpointConfig
    display_config_class = EmployerEmployeeDisplayConfig

    ordering = ["employer__computed_str", "id"]
    ordering_fields = ["employer__computed_str"]
    search_fields = ["employer__computed_str"]
    filterset_fields = {"employer": ["exact"]}

    def get_queryset(self):
        return super().get_queryset().filter(employee=self.kwargs["employee_id"])


class SystemEmployeeModelViewSet(EmployeeEmployerModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/systememployee.md"

    def __init__(self, **kwargs):
        global_preferences = global_preferences_registry.manager()
        employer_id = global_preferences["directory__main_company"]
        self.kwargs["employer_id"] = employer_id
        super().__init__(**kwargs)


class ClientManagerRelationshipRepresentationViewSet(RepresentationViewSet):
    serializer_class = ClientManagerRelationshipRepresentationSerializer
    queryset = ClientManagerRelationship.objects.all()


class ClientManagerViewSet(RevisionMixin, ModelViewSet):
    LIST_DOCUMENTATION = "directory/markdown/documentation/clientmanagerrelationship.md"
    filter_backends = (
        DjangoFilterBackend,
        filters.OrderingFilter,
        RankedFuzzySearchFilter,
    )
    min_rank = 0.25
    search_fields = ("client__computed_str", "relationship_manager__computed_str")
    ordering_fields = ("client__computed_str", "created", "relationship_manager__computed_str")
    ordering = ("-created", "id")
    queryset = ClientManagerRelationship.objects.none()
    serializer_class = ClientManagerModelSerializer
    display_config_class = ClientManagerModelDisplay
    endpoint_config_class = ClientManagerEndpoint
    filterset_class = ClientManagerFilter
    title_config_class = ClientManagerTitleConfig

    @cached_property
    def is_modifiable(self):
        if "pk" in self.kwargs and (obj := self.get_object()):
            return obj.status == ClientManagerRelationship.Status.DRAFT
        return self.new_mode

    def get_queryset(self):
        user = self.request.user
        profile = user.profile

        if user.is_superuser or profile.is_internal:  # TODO: Can TPM have access to this view?
            return ClientManagerRelationship.objects.select_related("client", "relationship_manager")
        return ClientManagerRelationship.objects.none()


class UserIsClientViewSet(ReadOnlyModelViewSet):
    queryset = ClientManagerRelationship.objects.all()
    LIST_DOCUMENTATION = "directory/markdown/documentation/userisclient.md"
    display_config_class = UserIsClientDisplayConfig
    title_config_class = UserIsClientTitleConfig
    endpoint_config_class = UserIsClientEndpointConfig
    serializer_class = UserIsClientModelSerializer
    ordering = ["-primary", "id"]
    ordering_fields = (
        "relationship_manager_name",
        "relationship_manager_email",
        "relationship_manager_phone_number",
    )

    permission_classes = [IsAuthenticated]

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(
                client_id=self.request.user.profile.id,
                status__in=[ClientManagerRelationship.Status.APPROVED, ClientManagerRelationship.Status.PENDINGREMOVE],
            )
            .annotate(
                relationship_manager_name=Concat(
                    F("relationship_manager__first_name"), Value(" "), F("relationship_manager__last_name")
                ),
                relationship_manager_email=Subquery(
                    EmailContact.objects.filter(primary=True, entry=OuterRef("relationship_manager")).values(
                        "address"
                    )[:1],
                    output_field=CharField(),
                ),
                relationship_manager_phone_number=Subquery(
                    TelephoneContact.objects.filter(primary=True, entry=OuterRef("relationship_manager")).values(
                        "number"
                    )[:1],
                    output_field=CharField(),
                ),
            )
        )



from unittest.mock import patch

import pytest


@pytest.mark.django_db
class TestScriptStep:
    @patch("wbcore.contrib.workflow.models.step.Step.set_failed")
    @patch("wbcore.contrib.workflow.models.step.Step.execute_single_next_step")
    def test_run(self, mock_execute, mock_failed, script_step_factory, process_step_factory):
        step = script_step_factory()
        process_step = process_step_factory(step=step)
        step.run(process_step)
        assert not mock_failed.called
        assert mock_execute.call_args.args == (process_step,)

    @patch("wbcore.contrib.workflow.models.step.Step.set_failed")
    @patch("wbcore.contrib.workflow.models.step.Step.execute_single_next_step")
    def test_run_failed(self, mock_execute, mock_failed, script_step_factory, process_step_factory):
        step = script_step_factory(script="Test")
        process_step = process_step_factory(step=step)
        step.run(process_step)
        assert mock_failed.call_args.args[0] == process_step
        assert not mock_execute.called



from django.contrib.contenttypes.models import ContentType
from django.dispatch import receiver
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy
from rest_framework.reverse import reverse
from reversion import is_registered
from reversion.models import Version

from wbcore import serializers
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display.instance_display import create_simple_display
from wbcore.signals.instance_buttons import add_extra_button

from ..serializers import VersionRepresentationSerializer


class VersionButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self):
        if self.request.user.is_staff or self.request.user.is_superuser:
            if (content_type_id := self.view.request.GET.get("content_type", None)) and (
                object_id := self.view.request.GET.get("object_id", None)
            ):
                content_type = ContentType.objects.get(id=content_type_id)
                return {
                    bt.HyperlinkButton(
                        label=_("Compare in Admin"),
                        endpoint=f"/admin/{content_type.app_label}/{content_type.model}/{object_id}/history/",
                    )
                }
        return set()

    def get_custom_list_instance_buttons(self):
        if (content_type_id := self.view.request.GET.get("content_type", None)) and (
            object_id := self.view.request.GET.get("object_id", None)
        ):
            content_type = ContentType.objects.get(id=content_type_id)
            obj = content_type.model_class().objects.get(id=object_id)

            class CompareSerializer(serializers.Serializer):
                compare_with = serializers.PrimaryKeyRelatedField(
                    label=gettext_lazy("Start"), queryset=Version.objects.get_for_object(obj)
                )
                _compare_with = VersionRepresentationSerializer(
                    source="compare_with", filter_params={"content_type": content_type_id, "object_id": object_id}
                )

            return {
                bt.ActionButton(
                    method=RequestType.GET,
                    identifiers=("{0.app_label}:{0.model}".format(content_type),),
                    key="revert",
                    label=_("Revert"),
                    icon=WBIcon.SYNCHRONIZE.icon,
                    description_fields=_(
                        """
                    <p>Revert object to this version</p>
                    """
                    ),
                    action_label=_("Reverting"),
                    title=_("Revert"),
                ),
                bt.ActionButton(
                    method=RequestType.GET,
                    identifiers=("wbcore:version",),
                    key="compare_with",
                    label=_("Compare With"),
                    icon=WBIcon.DATA_GRID.icon,
                    description_fields=_(
                        """
                    <p>Compare this version with another version</p>
                    """
                    ),
                    serializer=CompareSerializer,
                    action_label=_("Comparing"),
                    title=_("Compare With"),
                    instance_display=create_simple_display([["compare_with"]]),
                ),
            }
        return set()

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()


@receiver(add_extra_button)
def add_reversion_documents(sender, instance, request, view, pk=None, **kwargs):
    if instance and pk and view and is_registered(view.get_model()):
        content_type = view.get_content_type()
        endpoint = (
            f'{reverse("wbcore:version-list", args=[], request=request)}?content_type={content_type.id}&object_id={pk}'
        )
        return bt.WidgetButton(endpoint=endpoint, label="History", icon=WBIcon.HISTORY.icon)



from psycopg.types.range import NumericRange
from rest_framework import serializers

from .mixins import RangeMixin, WBCoreSerializerFieldMixin, decorator
from .types import DisplayMode, WBCoreType

percent_decorator = decorator(position="right", value="%", decorator_type="text")


class NumberFieldMixin:
    def __init__(
        self,
        *args,
        percent=False,
        display_mode=DisplayMode.DECIMAL,
        precision=2,
        signed: bool = True,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.percent = percent
        self.disable_formatting = getattr(self, "disable_formatting", kwargs.pop("disable_formatting", False))
        self.display_mode = display_mode
        self.precision = kwargs.get("decimal_places", precision)
        self.max_digits = kwargs.get("max_digits", 34)
        self.signed = signed

    def set_precision(self, precision):
        self.precision = precision
        self._kwargs["precision"] = self.precision

    def set_percent(self, percent):
        self.percent = percent
        self._kwargs["percent"] = self.percent

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["display_mode"] = self.display_mode.value
        representation["precision"] = self.precision
        representation["max_digits"] = self.max_digits
        representation["disable_formatting"] = self.disable_formatting
        representation["signed"] = self.signed

        if self.percent:  # TODO: Discuss with Christoph if this is necessary like this
            representation["type"] = WBCoreType.PERCENT.value
            representation["precision"] = self.precision - 2
        return key, representation


class IntegerField(NumberFieldMixin, WBCoreSerializerFieldMixin, serializers.IntegerField):
    field_type = WBCoreType.NUMBER.value

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.precision = 0


class YearField(IntegerField):
    disable_formatting = True


class DecimalField(NumberFieldMixin, WBCoreSerializerFieldMixin, serializers.DecimalField):
    field_type = WBCoreType.NUMBER.value

    # TODO: If this is used, then the validation for max_digits and decimal_fields is not done
    # def validate_precision(self, value):
    #     return value

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        field_name, representation = super().get_representation(request, field_name)

        if meta := getattr(self.parent, "Meta", None):
            odp = getattr(meta, "override_decimal_places", None)
            percent_fields = getattr(meta, "percent_fields", [])
            if odp is not None and field_name not in percent_fields:
                representation["precision"] = odp

        return field_name, representation


class FloatField(NumberFieldMixin, WBCoreSerializerFieldMixin, serializers.FloatField):
    field_type = WBCoreType.NUMBER.value


class DecimalRangeField(RangeMixin, WBCoreSerializerFieldMixin, serializers.DecimalField):
    field_type = WBCoreType.NUMBERRANGE.value
    internal_field = NumericRange

    def __init__(self, max_digits=None, decimal_places=None, **kwargs):
        if not max_digits:
            max_digits = 3
        if not decimal_places:
            decimal_places = 2
        super().__init__(max_digits, decimal_places, **kwargs)



import logging
from typing import Any, Callable, Generic, TypeVar

from celery import shared_task
from django.db import models
from django.db.models.signals import ModelSignal
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import BaseMessage
from langchain_openai import ChatOpenAI

from ..exceptions import APIStatusErrors, BadRequestErrors
from .utils import run_llm

logger = logging.getLogger("llm")
from pydantic import BaseModel


@shared_task(
    queue="llm",
    autoretry_for=tuple(APIStatusErrors),
    retry_backoff=10,
    max_retries=5,  # retry 5 times maximum
    default_retry_delay=60,  # retry in 60s
    retry_jitter=False,
)
def invoke_as_task(
    instance,
    prompt: str,
    chat_model: type[BaseChatModel],
    chat_model_name: str,
    max_tokens: int,
    model_field: str | None,
    output_model: "type[BaseModel] | None",
    tools: list[dict[str, Any]],
    query: Callable | dict[str, Any] | None,
    extra_query: dict[str, Any] | None,
    llm_kwargs: dict[str, Any] | None,
    result_parser: Callable | None,
):
    try:
        if callable(query):
            query = query(instance)
        if not query:
            query = dict()
        if extra_query:
            query.update(extra_query)

        output_result, ai_msg = run_llm(
            prompt, output_model, chat_model, chat_model_name, max_tokens, query=query, extra_tools=tools, **llm_kwargs
        )
        # if a result parser is provided, we use this to parse the results into the instance
        if result_parser:
            instance = result_parser(instance, output_result, ai_msg)
        elif output_result and isinstance(output_result, BaseModel):
            for field, value in output_result.model_dump().items():
                setattr(instance, field, value)
    except tuple(BadRequestErrors) as e:  # we silent bad request error because there is nothing we can do about it
        logger.warning(str(e))
    except tuple(APIStatusErrors) as e:  # for APIStatusError, we let celery retry it
        raise e
    except Exception as e:  # otherwise we log the error and silently fail
        logger.warning(str(e))
    return instance


T = TypeVar("T", bound=models.Model)


add_llm_prompt = ModelSignal()


class LLMConfig(Generic[T]):
    def __init__(
        self,
        key: str,
        prompt: Callable | str | list[BaseMessage],
        field: str | None = None,
        output_model: "type[BaseModel] | None" = None,
        on_save: bool = True,
        on_condition: Callable | bool | None = None,
        chat_model: Callable | tuple[type[BaseChatModel], str] = (ChatOpenAI, "gpt-4o-mini"),
        max_tokens: int = 16000,
        tools: Callable | list[dict[str, str]] | None = None,
        query: Callable | dict[str, Any] | None = None,
        llm_kwargs_callback: Callable | None = None,
        result_parser: Callable | None = None,
        **llm_kwargs,
    ):
        self.key = key
        self.on_save = on_save
        self.field = field
        self.on_condition = on_condition
        self.prompt = prompt
        self.output_model = output_model
        self.chat_model = chat_model
        self.max_tokens = max_tokens
        self.tools = tools
        self.query = query
        self.llm_kwargs_callback = llm_kwargs_callback
        self.llm_kwargs = llm_kwargs
        self.result_parser = result_parser

    def check_condition(self, instance: T) -> bool:
        if self.on_condition is None:
            return True
        if callable(self.on_condition):
            return self.on_condition(instance)
        return bool(self.on_condition)

    def _get_prompt(self, instance: T):
        prompt = self.prompt
        extra_query = dict()
        if callable(prompt):
            prompt = prompt(instance)
        for _, response in add_llm_prompt.send(sender=instance.__class__, instance=instance, key=self.key):
            remote_prompts, remote_query = response
            prompt.extend(remote_prompts)
            extra_query.update(remote_query)

        return prompt, extra_query

    def _get_chat_model(self, instance: T) -> tuple[type[BaseChatModel], str]:
        if callable(self.chat_model):
            return self.chat_model(instance)
        return self.chat_model

    def _get_tools(self, instance: T) -> list[dict[str, Any]]:
        if callable(self.tools):
            return self.tools(instance)
        return self.tools

    def _get_llm_kwargs(self, instance: T) -> dict[str, Any]:
        llm_kwargs = self.llm_kwargs
        if callable(self.llm_kwargs_callback):
            llm_kwargs = self.llm_kwargs_callback(instance)
        return llm_kwargs

    def schedule(self, instance: T, initial: bool = True):
        prompt, extra_query = self._get_prompt(instance)
        args = []
        if initial:
            args.append(instance)
        chat_model, chat_model_name = self._get_chat_model(instance)
        tools = self._get_tools(instance)
        llm_kwargs = self._get_llm_kwargs(instance)
        args.extend(
            [
                prompt,
                chat_model,
                chat_model_name,
                self.max_tokens,
                self.field,
                self.output_model,
                tools,
                self.query,
                extra_query,
                llm_kwargs,
                self.result_parser,
            ]
        )
        return invoke_as_task.s(*args)

    def invoke(self, instance: T):
        self.schedule(instance)()
